#!/usr/bin/perl -w

use strict;

#SCORING PARAMETERS:=================================================

# min_head_overlap is the minimum mutual fractional overlap allowed
# for a mention to be declared as successfully detected.
    use vars qw ($min_head_overlap);
    $min_head_overlap = 0.3; #minimum fractional overlap for mention detection

# min_name_overlap is the minimum mutual fractional overlap allowed
# for a name to be declared as successfully detected.
    use vars qw ($min_name_overlap);
    $min_name_overlap = 0.3; #minimum fractional overlap for name detection

# maa_chars is the mis-alignment allowance used in calculating span overlap.
# (the actual overlap is augmented by maa_chars)
    use vars qw ($maa_chars);
    $maa_chars = 1;

# max_name_diff_chars is the maximum difference allowed between the end points
# of two names in order for them to be declared a "match".
    use vars qw ($max_name_diff_chars);
    $max_name_diff_chars = 4;

# max_mention_diff_chars is the maximum difference allowed between the end points
# of two mentions in order for them to be declared a "match".
    use vars qw ($max_mention_diff_chars);
    $max_mention_diff_chars = 4;

#The following three parameters are used to map ref entities onto tst entities
# nam_match is the weight assigned to mentions of type NAME
# nom_match is the weight assigned to mentions of type NOMINAL
# pro_match is the weight assigned to mentions of type PRONOUN
use vars qw ($nam_match $nom_match $pro_match);
    $nam_match = 1.00;
    $nom_match = 0.20;
    $pro_match = 0.04;

#====================================================================

#################################
# History:
#
# version 12
#    did some cosmetic cleanup in output
#    added parameter printout
#    added extent alignment in printout
#
# version 13
#    added optional ref-tst entity pair output
#
# version 14
#    provided diagnostic output using source data
#
# version 15
#    provided html file-based comparison of ref with tst
#
# version 16
#    sgml tags added to html display
#
# version 17
#    added check/notification for mention type differences
#    html comparison output conditioned on output control flag
#        (-x outputs only entities that are in error)
#    changed html comparison output to output all data for a
#        single document in one file
#    changed the html color sequence for entities so that the first
#        color used is red (rather than blue) to facilitate visualization
#
# version 18
#    added error percentage computation to the error statistics report
#    added underlining of mentions in html output for unmapped entities
#
# version 19
#    added yellow background to those characters in the html output for which
#    the different annotations disagree.  This makes the differences much more
#    visually apparent.
#
# version 20
#    expanded the scoring output to condition the evaluation of entity detection
#    on the entity's type and number of mentions.
#
# version 21
#    inserted an entity_mapping_score table to avoid repeated computation of
#    entity mapping scores and thus speed up computation, especially for the
#    computationally burdensome cases where the cohort set is large.
#
# version 22
#    rearranged entity mapping to skip comparisons for which a better match
#    is impossible.  This is performed by comparing the score for the current
#    best mapping with the maximum possible score for the current (partial)
#    mapping and abandoning the search whenever the maximum possible score
#    would be worse than the current best score.  More rapid search is
#    achieved by sorting the entities according to maximum possible score
#    and putting the most important entities in the innermost loops.
#
# version 23
#    bug fix to reinstate ability to output html visualization files.
#
# version 24
#    bug fix to correct the count of the number of mentions with correctly
#    recognized extents.
#
#################################

use vars qw ($print_entities $print_all_data $print_err_data);
use vars qw ($print_ent_data $print_men_data $print_nam_data);

use vars qw ($print_mapping_detail $print_timing);

use vars qw ($input_file $input_doc $input_entity);
use vars qw ($usage);
use vars qw ($best_score);
use vars qw (%entity_detection_statistics);
use vars qw (%entity_type_statistics);
use vars qw (%mention_detection_statistics);
use vars qw (%name_detection_statistics);
use vars qw (%entity_mapping_score);

use vars qw (@error_types @entity_types @mention_types @count_types);
@error_types = ("correct", "miss", "fa", "error");
@entity_types = ("FACILITY", "GPE", "LOCATION", "ORGANIZATION", "PERSON", "total");
@mention_types = ("NAME", "NOMINAL", "PRONOUN", "total");
@count_types = ("1", "2", "3-4", "5-8", ">8", "total");
#
#N.B.  The above values of entity_types, mention_types and count_types MUST BE ALL UNIQUE!!!
#      (This is because they are all keys in the SAME hash table.)
#

my $version = "24";

my ($date, $time) = date_time_stamp();
print "EDT reference comparison (version $version) run on $date at $time\n";
print "command line:  ", $0, " ", join(" ", @ARGV), "\n";
print "PARAMETERS:\n";
print "  $min_head_overlap = min mutual fractional overlap of matching mention heads\n";
print "  $min_name_overlap = min mutual fractional overlap of matching names\n";
print "  $maa_chars = \# of characters mis-alignment allowance in span overlap calculation\n";
print "  $max_name_diff_chars = max \# of chars difference in endpoints for names to match\n";
print "  $max_mention_diff_chars = max \# of chars difference in endpoints for mention extents to match\n";
print "  $nam_match/$nom_match/$pro_match = entity mapping weights for NAME/NOMINAL/PRONOUN mentions\n";

$usage = "\n\n$0 [-hsaxemn] -r <ref_list> -t <tst_list> -q <src_list>\n\n".
    "Description:  This Perl program compares two different EDT annotations\n".
    "              for a set of source files.  The comparison is in terms of\n".
    "              the EDT tasks, namely entity detection, mention detection,\n".
    "              attribute recognition, and recognition of mention extent.\n".
    "\n".
    "Required arguments:\n".
    "  -r <ref_list> is a file containing a list of EDT annotation output files.\n".
    "      These EDT annotation output files must be in standard EDT xml format.\n".
    "      The annotation specified by ref_list is treated as 'reference' data.\n".
    "  -t <tst_list> is a file containing a list of EDT annotation output files.\n".
    "      These EDT annotation output files must be in standard EDT xml format.\n".
    "      The annotation specified by tst_list is treated as 'test' data.\n".
    "  -q <src_list> is a file containing a list of source data files.\n".
    "      These files must be in EDT sgml format and contain document ID info.\n".
    "\n".
    "Optional arguments:\n".
    "  -h prints this help message to STDOUT \n".
    "  -s prints EDT annotation data summary to STDOUT\n".
    "  -p <map_file> writes ref-tst entity pairs to an entity-pair map file.\n".
    "      Each record contains one ref-tst entity pair.  Unmapped entities are not\n".
    "      output.  The output format is quoted entity ID's, white-space separated:\n".
    "             \"ref-ID\" \"tst-ID\"\n".
    "  -c <cmp_dir> writes an html file to the directory cmp_dir for each document.\n".
    "      These *.html files provide a visual illustration of mapped ref and tst\n".
    "      entities, in terms of their mentions.  ref and tst are displayed parallel\n".
    "      to each other.  Differences are highlighed to aid reconciliation of entity\n".
    "      annotation.\n".
    "\n".
    "      The following flags control the output of comparison data for diagnosis:\n".
    "      There are two comparison outputs, a printed \"report\" of ref/tst differences\n".
    "      sent to STDOUT and an html \"illustration\" of entity mentions written to\n".
    "      directory <cmp_dir>, one file per document.  The html file name is\n".
    "      \"EDT_COMP_<doc_id>.html\"\n".
    "  -a outputs report data and html illustrations for all entities\n".
    "  -x outputs report data and html illustrations for entities with errors\n".
    "  -e outputs report data for entities with entity detection/type errors\n".
    "  -m outputs report data for entities with mention detection/extent errors\n".
    "  -n outputs report data for entities with name detection/extent errors\n".
    "\n".
    "      The following flags provide diagnostic output for program development:\n".
    "  -i outputs the time spent mapping ref entities to tst entities for each doc\n".
    "  -d outputs detailed mapping info and score for every mapping computed\n".
    "\n";

{
    use vars qw ($opt_h $opt_s $opt_a $opt_x $opt_e $opt_m $opt_n $opt_c);
    use vars qw ($opt_p $opt_q $opt_r $opt_t $opt_i $opt_d);
    require "getopts.pl";
    &Getopts ('hsaxemnidc:p:q:r:t:');
    die $usage if defined($opt_h);
    $print_entities = defined($opt_s) ? 1 : 0;
    $print_all_data = defined($opt_a) ? 1 : 0;
    $print_err_data = defined($opt_x) ? 1 : 0;
    $print_ent_data = defined($opt_e) ? 1 : 0;
    $print_men_data = defined($opt_m) ? 1 : 0;
    $print_nam_data = defined($opt_n) ? 1 : 0;
    $print_timing = defined($opt_i) ? 1 : 0;
    $print_mapping_detail = defined($opt_d) ? 1 : 0;
    die "Error in command line$usage"
	unless (defined($opt_q) and defined($opt_r) and defined($opt_t));

    my ($source_list) = $opt_q;
    my ($list0) = $opt_r;
    my ($list1) = $opt_t;
    my (%ref_docs0, %ref_docs1);

#read in all EDT reference data from files in list 0
    open (LIST, $list0) or die "\nUnable to open file list '$list0'", $usage;
    while ($input_file = <LIST>) {
	get_EDT_ref_data (\%ref_docs0, $input_file);
    }
    close (LIST);
    
#read in all EDT reference data from files in list 1
    open (LIST, $list1) or die "\nUnable to open file list '$list1'", $usage;
    while ($input_file = <LIST>) {
	get_EDT_ref_data (\%ref_docs1, $input_file);
    }
    close (LIST);
    
#read in all EDT source data from files in the source list
    open (LIST, $source_list) or die "\nUnable to open file list '$source_list'", $usage;
    while ($input_file = <LIST>) {
	get_EDT_source_data (\%ref_docs0, $input_file);
    }
    close (LIST);

#print input data
    if ($print_entities) {
	print "REF ", "=" x 30, " REF ", "=" x 30, " REF\n";
	print_EDT_ref_data (%ref_docs0);
	print "END ", "=" x 30, " END ", "=" x 30, " END\n";
	print "TST ", "=" x 30, " TST ", "=" x 30, " TST\n";
	print_EDT_ref_data (%ref_docs1);
	print "END ", "=" x 30, " END ", "=" x 30, " END\n";
    }
    
#compare input data and evaluate differences
    compare_EDT_refs (\%ref_docs0, \%ref_docs1);

#output ref-tst entity pairs
    output_entity_pairs ($opt_p, \%ref_docs0) if defined($opt_p);

#output scoring details
    output_entity_comparisons ($opt_c, \%ref_docs0, \%ref_docs1) if defined ($opt_c);
}    
    
#################################

sub compare_EDT_refs {

    my ($ref, $tst) = @_;
    my $doc_id;
    my ($ref_doc, $tst_doc);
    my ($ref_entities, $tst_entities);

    initialize_statistics ();

    foreach $doc_id (keys %$ref) {

	$ref_doc = $ref->{$doc_id};
	$tst_doc = $tst->{$doc_id};
	die "\n\nFATAL ERROR:  ref doc '$doc_id' has no corresponding test doc\n\n"
	    unless  defined($tst_doc);

#map the entities and mentions
	$ref_entities = $ref_doc->{entities};
	$tst_entities = $tst_doc->{entities};
	undef(%entity_mapping_score); #N.B. -- this assumes no cross-doc entities
	my $tstart = (times)[0];
	map_EDT_entities ($ref_entities, $tst_entities);
	if ($print_timing) {
	    my $mapping_time = (times)[0] - $tstart;
	    my $nr = @$ref_entities;
	    my $nrm = 0;
	    foreach my $entity (@$ref_entities) {
		$nrm += @{$entity->{mentions}};
	    }
	    my $nt = @$tst_entities;
	    my $ntm = 0;
	    foreach my $entity (@$tst_entities) {
		$ntm += @{$entity->{mentions}};
	    }
	    printf "%10.2f secs to map%3d ref entities (%3d mentions) to%3d tst entities (%3d mentions) for doc %s\n",
	        $mapping_time, $nr, $nrm, $nt, $ntm, $doc_id;
	}

#tabulate scoring statistics
	tabulate_EDT_scoring_statistics ($ref_entities, $tst_entities);
	print_scoring_details ($ref_entities, $tst_entities)
	    if ($print_all_data or $print_err_data or $print_ent_data
		or $print_men_data or $print_nam_data);
    }

#score entity detection
    score_EDT_entity_detection ();
    
#score attribute recognition
    score_EDT_attribute_recognition ();

#score mention recognition
    score_EDT_mention_recognition ();
}

#################################

sub score_EDT_entity_detection {

    my $type;

#scoring conditioned on entity type
    foreach $type (@entity_types) {
	next if $type eq "total";
	$entity_detection_statistics{total}{miss} += $entity_detection_statistics{$type}{miss};
	$entity_detection_statistics{total}{fa} += $entity_detection_statistics{$type}{fa};
	$entity_detection_statistics{total}{correct} += $entity_detection_statistics{$type}{correct};
	$entity_detection_statistics{total}{error} += $entity_detection_statistics{$type}{error};
    }

    print "\nEntity Detection and Type Recognition statistics:\n",
    " ref     ____________count______________       ____________percent____________\n",
    " entity  Detection      Type_Recognition       Detection      Type_Recognition\n",
    " type    miss    fa     miss   err  corr       miss    fa     miss   err  corr\n";
    foreach $type (@entity_types) {
	my $total = ($entity_detection_statistics{$type}{miss} +
		     $entity_detection_statistics{$type}{error} +
		     $entity_detection_statistics{$type}{correct});
	printf "  %3.3s%8d%6d%9d%6d%6d%11.1f%6.1f%9.1f%6.1f%6.1f\n", $type,
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{fa},
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{error},
	$entity_detection_statistics{$type}{correct},
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{fa}/max($total,1),
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{error}/max($total,1),
	100*$entity_detection_statistics{$type}{correct}/max($total,1);
    }

#scoring conditioned on mention type
    $entity_detection_statistics{total}{miss} =
	$entity_detection_statistics{total}{fa} =
	    $entity_detection_statistics{total}{correct} =
		$entity_detection_statistics{total}{error} = 0;
	
    foreach $type (@mention_types) {
	next if $type eq "total";
	$entity_detection_statistics{total}{miss} += $entity_detection_statistics{$type}{miss};
	$entity_detection_statistics{total}{fa} += $entity_detection_statistics{$type}{fa};
	$entity_detection_statistics{total}{correct} += $entity_detection_statistics{$type}{correct};
	$entity_detection_statistics{total}{error} += $entity_detection_statistics{$type}{error};
    }

    print "\nEntity Detection and Type Recognition statistics:\n",
    " ref     ____________count______________       ____________percent____________\n",
    " mention Detection      Type_Recognition       Detection      Type_Recognition\n",
    " type    miss    fa     miss   err  corr       miss    fa     miss   err  corr\n";
    foreach $type (@mention_types) {
	my $total = ($entity_detection_statistics{$type}{miss} +
		     $entity_detection_statistics{$type}{error} +
		     $entity_detection_statistics{$type}{correct});
	printf "  %3.3s%8d%6d%9d%6d%6d%11.1f%6.1f%9.1f%6.1f%6.1f\n", $type,
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{fa},
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{error},
	$entity_detection_statistics{$type}{correct},
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{fa}/max($total,1),
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{error}/max($total,1),
	100*$entity_detection_statistics{$type}{correct}/max($total,1);
    }

#scoring conditioned on count type
    $entity_detection_statistics{total}{miss} =
	$entity_detection_statistics{total}{fa} =
	    $entity_detection_statistics{total}{correct} =
		$entity_detection_statistics{total}{error} = 0;
	
    foreach $type (@count_types) {
	next if $type eq "total";
	$entity_detection_statistics{total}{miss} += $entity_detection_statistics{$type}{miss};
	$entity_detection_statistics{total}{fa} += $entity_detection_statistics{$type}{fa};
	$entity_detection_statistics{total}{correct} += $entity_detection_statistics{$type}{correct};
	$entity_detection_statistics{total}{error} += $entity_detection_statistics{$type}{error};
    }

    print "\nEntity Detection and Type Recognition statistics:\n",
    " ref     ____________count______________       ____________percent____________\n",
    " mention Detection      Type_Recognition       Detection      Type_Recognition\n",
    " count   miss    fa     miss   err  corr       miss    fa     miss   err  corr\n";
    foreach $type (@count_types) {
	my $total = ($entity_detection_statistics{$type}{miss} +
		     $entity_detection_statistics{$type}{error} +
		     $entity_detection_statistics{$type}{correct});
	printf "  %3.3s%8d%6d%9d%6d%6d%11.1f%6.1f%9.1f%6.1f%6.1f\n", $type,
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{fa},
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{error},
	$entity_detection_statistics{$type}{correct},
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{fa}/max($total,1),
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{error}/max($total,1),
	100*$entity_detection_statistics{$type}{correct}/max($total,1);
    }
}

#################################

sub score_EDT_attribute_recognition {

    my ($type, $ref_type);

    foreach $ref_type (keys %entity_type_statistics) {
	foreach $type (keys %entity_type_statistics) {
	    next if $type eq "total";
	    $entity_type_statistics{$ref_type}{total} += $entity_type_statistics{$ref_type}{$type};
	}
    }

    print "\nEntity Type confusion matrix (for mapped entities):\n",
    "               ___________count___________        __________percent__________\n",
    "  ref \\ tst:  ";
    foreach $type (@entity_types) {
	last if $type eq "total";
	printf " %3.3s  ", $type;
    }
    print "   ";
    foreach $type (@entity_types) {
	last if $type eq "total";
	printf "   %3.3s", $type;
    }
    print "\n";
    foreach $ref_type (@entity_types) {
	last if $ref_type eq "total";
	printf "  %3.3s       ", $ref_type;
	foreach $type (@entity_types) {
	    printf "%6d", $entity_type_statistics{$ref_type}{$type}
	    unless $type eq "total";
	}
	print "     ";
	foreach $type (@entity_types) {
	    printf "%6.1f", 100*$entity_type_statistics{$ref_type}{$type} /
		max($entity_type_statistics{$ref_type}{total},1)
		    unless $type eq "total";
	}
	print "\n";
    }

    foreach $type (keys %name_detection_statistics) {
	next if $type eq "total";
	$name_detection_statistics{total}{miss} += $name_detection_statistics{$type}{miss};
	$name_detection_statistics{total}{fa} += $name_detection_statistics{$type}{fa};
	$name_detection_statistics{total}{correct} += $name_detection_statistics{$type}{correct};
	$name_detection_statistics{total}{error} += $name_detection_statistics{$type}{error};
    }

    print "\nName Detection and Extent Recognition statistics (for mapped entities):\n",
    " ref     ____________count______________       ____________percent____________\n",
    " entity  Detection     Extent_Recognition      Detection     Extent_Recognition\n",
    " type    miss    fa     miss   err  corr       miss    fa     miss   err  corr\n";
    foreach $type (@entity_types) {
	my $total = ($name_detection_statistics{$type}{miss} +
		     $name_detection_statistics{$type}{error} +
		     $name_detection_statistics{$type}{correct});
	printf "  %3.3s%8d%6d%9d%6d%6d%11.1f%6.1f%9.1f%6.1f%6.1f\n", $type,
	$name_detection_statistics{$type}{miss},
	$name_detection_statistics{$type}{fa},
	$name_detection_statistics{$type}{miss},
	$name_detection_statistics{$type}{error},
	$name_detection_statistics{$type}{correct},
	100*$name_detection_statistics{$type}{miss}/max($total,1),
	100*$name_detection_statistics{$type}{fa}/max($total,1),
	100*$name_detection_statistics{$type}{miss}/max($total,1),
	100*$name_detection_statistics{$type}{error}/max($total,1),
	100*$name_detection_statistics{$type}{correct}/max($total,1);
    }
}

#################################

sub score_EDT_mention_recognition {

    my $type;

    foreach $type (keys %mention_detection_statistics) {
	next if $type eq "total";
	$mention_detection_statistics{total}{miss} += $mention_detection_statistics{$type}{miss};
	$mention_detection_statistics{total}{fa} += $mention_detection_statistics{$type}{fa};
	$mention_detection_statistics{total}{correct} += $mention_detection_statistics{$type}{correct};
	$mention_detection_statistics{total}{error} += $mention_detection_statistics{$type}{error};
    }

    print "\nMention Detection and Extent Recognition statistics (for mapped entities):\n",
    " ref     ____________count______________       ____________percent____________\n",
    " mention Detection     Extent_Recognition      Detection     Extent_Recognition\n",
    " type    miss    fa     miss   err  corr       miss    fa     miss   err  corr\n";
    foreach $type (@mention_types) {
	my $total = ($mention_detection_statistics{$type}{miss} +
		     $mention_detection_statistics{$type}{error} +
		     $mention_detection_statistics{$type}{correct});
	printf "  %3.3s%8d%6d%9d%6d%6d%11.1f%6.1f%9.1f%6.1f%6.1f\n", $type,
	$mention_detection_statistics{$type}{miss},
	$mention_detection_statistics{$type}{fa},
	$mention_detection_statistics{$type}{miss},
	$mention_detection_statistics{$type}{error},
	$mention_detection_statistics{$type}{correct},
	100*$mention_detection_statistics{$type}{miss}/max($total,1),
	100*$mention_detection_statistics{$type}{fa}/max($total,1),
	100*$mention_detection_statistics{$type}{miss}/max($total,1),
	100*$mention_detection_statistics{$type}{error}/max($total,1),
	100*$mention_detection_statistics{$type}{correct}/max($total,1);
    }
}

#################################

sub entity_mention_type {

    my ($ref_entity) = @_;
    my ($ref_mentions, $ref_mention);

#determine the type of mentions for this entity
    my $mention_type = "PRONOUN";
    $ref_mentions = $ref_entity->{mentions};
    foreach $ref_mention (@$ref_mentions) {
	$mention_type = "NAME" if ($ref_mention->{TYPE} eq "NAME");
	$mention_type = $ref_mention->{TYPE} if $mention_type eq "PRONOUN";
    }
    return $mention_type;
}

#################################

sub entity_count_type {

    my ($ref_entity) = @_;
    my ($count, $count_type);

#determine the number of mentions for this entity
    $count = @{$ref_entity->{mentions}};
    return $count le 1 ? $count_types[0] :
	   $count le 2 ? $count_types[1] :
	   $count le 4 ? $count_types[2] :
	   $count le 8 ? $count_types[3] : $count_types[4];
}

#################################

sub tabulate_EDT_scoring_statistics {

    my ($ref_entities, $tst_entities) = @_;
    my ($ref_entity, $tst_entity);
    my ($ref_mentions, $tst_mentions);
    my ($ref_mention, $tst_mention);
    my ($ref_names, $tst_names);
    my ($ref_name, $tst_name);
    my ($overlap, $max_overlap);

#accumulate recognition statistics over all entities
    foreach $ref_entity (@$ref_entities) {

	my $mention_type = entity_mention_type ($ref_entity);
	my $count_type = entity_count_type ($ref_entity);

#measure performance only for those entities that are correctly detected
	if ($ref_entity->{MAP}) {

#update entity detection statistics
	    if ($ref_entity->{TYPE} eq $ref_entity->{MAP}{TYPE}) {
		$entity_detection_statistics{$ref_entity->{TYPE}}{correct}++;
		$entity_detection_statistics{$mention_type}{correct}++;
		$entity_detection_statistics{$count_type}{correct}++;
	    }
	    else {
		$entity_detection_statistics{$ref_entity->{TYPE}}{error}++;
		$entity_detection_statistics{$mention_type}{error}++;
		$entity_detection_statistics{$count_type}{error}++;
	    }

#update entity type statistics
	    $entity_type_statistics{$ref_entity->{TYPE}}{$ref_entity->{MAP}{TYPE}}++;

#update entity name statistics
	    $ref_names = $ref_entity->{names};
	    $tst_names = $ref_entity->{MAP}{names};
	    foreach $ref_name (@$ref_names) {
		$max_overlap = $min_name_overlap;
		foreach $tst_name (@$tst_names) {
		    $overlap = span_overlap($ref_name, $tst_name);
		    if ($overlap > $max_overlap) {
			$max_overlap = $overlap;
			$ref_name->{MAP} = $tst_name;
		    }
		}
		if ($ref_name->{MAP}) {
		    $tst_name = $ref_name->{MAP};
		    $tst_name->{MAP} = $ref_name;
		    span_endpoint_mismatch($ref_name, $tst_name)
			<= $max_name_diff_chars ?
			    $name_detection_statistics{$ref_entity->{TYPE}}{correct}++ :
				$name_detection_statistics{$ref_entity->{TYPE}}{error}++;
		}
		else {
		    $name_detection_statistics{$ref_entity->{TYPE}}{miss}++;
		}
	    }
	    foreach $tst_name (@$tst_names) {
		if (not $tst_name->{MAP}) {
		    $name_detection_statistics{$ref_entity->{TYPE}}{fa}++;
		}
	    }

#update mention statistics
	    $ref_mentions = $ref_entity->{mentions};
	    $tst_mentions = $ref_entity->{MAP}->{mentions};
	    foreach $ref_mention (@$ref_mentions) {
		foreach $tst_mention (@$tst_mentions) {
		    if (mention_mapping_score($ref_mention, $tst_mention)) {
			$ref_mention->{MAP} = $tst_mention;
			$tst_mention->{MAP} = $ref_mention;
		    }
		}
		if ($ref_mention->{MAP}) {
		    span_endpoint_mismatch
			($ref_mention->{extent}, $ref_mention->{MAP}->{extent})
			    <= $max_mention_diff_chars ?
				$mention_detection_statistics{$ref_mention->{TYPE}}{correct}++ :
				    $mention_detection_statistics{$ref_mention->{TYPE}}{error}++;
		}
		else {
		    $mention_detection_statistics{$ref_mention->{TYPE}}{miss}++;
		}
	    }
	    foreach $tst_mention (@$tst_mentions) {
		if (not $tst_mention->{MAP}) {
		    $mention_detection_statistics{$tst_mention->{TYPE}}{fa}++;
		}
	    }
	}
	else {
	    $entity_detection_statistics{$ref_entity->{TYPE}}{miss}++;
	    $entity_detection_statistics{$mention_type}{miss}++;
	    $entity_detection_statistics{$count_type}{miss}++;
	}
    }
    foreach $tst_entity (@$tst_entities) {
	if (not $tst_entity->{MAP}) {
	    $entity_detection_statistics{$tst_entity->{TYPE}}{fa}++;
	    $entity_detection_statistics{entity_mention_type($tst_entity)}{fa}++;
	    $entity_detection_statistics{entity_count_type($tst_entity)}{fa}++;
	}
    }
}

#################################

sub initialize_statistics {

    my ($ref_type, $tst_type, $err_type, $men_type);

    foreach $ref_type (@entity_types) {
	foreach $err_type (@error_types) {
	    $entity_detection_statistics{$ref_type}{$err_type} = 0;
	    $name_detection_statistics{$ref_type}{$err_type} = 0;
	}
	foreach $tst_type (@entity_types) {
	    $entity_type_statistics{$ref_type}{$tst_type} = 0;
	}
    }
    foreach $ref_type (@count_types) {
	foreach $err_type (@error_types) {
	    $entity_detection_statistics{$ref_type}{$err_type} = 0;
	}
    }
    foreach $men_type (@mention_types) {
	foreach $err_type (@error_types) {
	    $mention_detection_statistics{$men_type}{$err_type} = 0;
	    $entity_detection_statistics{$men_type}{$err_type} = 0;
	}
    }
}

#################################

sub print_scoring_details {

    my ($ref_entities, $tst_entities) = @_;
    my ($ref_entity, $tst_entity);
    my ($ref_mentions, $tst_mentions);
    my ($ref_mention, $tst_mention);
    my ($ref_names, $tst_names);
    my ($ref_name, $tst_name);
    my $output;
    my $separator = "-" x 18 . "\n";
    my $print_data;
    my @pad;

    foreach $ref_entity (@$ref_entities) {
	$output = $separator;
	$print_data = $print_all_data;
	if ($ref_entity->{MAP}) {
	    if ($ref_entity->{TYPE} eq $ref_entity->{MAP}->{TYPE}) {
		$output .= "    ref entity $ref_entity->{ID} ($ref_entity->{TYPE})\n";
		$output .= "    tst entity $ref_entity->{MAP}->{ID} ($ref_entity->{MAP}->{TYPE})\n";
	    }
	    else {
		$print_data ||= $print_err_data || $print_ent_data;
		$output .= ">>> ref entity $ref_entity->{ID} ($ref_entity->{TYPE}) -- ENTITY TYPE MISMATCH\n";
		$output .= ">>> tst entity $ref_entity->{MAP}->{ID} ($ref_entity->{MAP}->{TYPE}) -- ENTITY TYPE MISMATCH\n";
	    }

	    $ref_mentions = $ref_entity->{mentions};
	    $tst_mentions = $ref_entity->{MAP}->{mentions};
	    foreach $ref_mention (@$ref_mentions) {
		if ($ref_mention->{MAP}) {
		    if ($ref_mention->{TYPE} eq $ref_mention->{MAP}->{TYPE}) {
			$output .= "      ref mention = '$ref_mention->{head}{text}'";
			$output .= " ($ref_mention->{head}{start}:$ref_mention->{head}{end})";
			$output .= " ($ref_mention->{TYPE})\n";
		    }
		    else {
			$print_data ||= $print_err_data || $print_men_data;
			$output .= ">>>   ref mention = '$ref_mention->{head}{text}'";
			$output .= " ($ref_mention->{head}{start}:$ref_mention->{head}{end})";
			$output .= " ($ref_mention->{TYPE}) -- MENTION TYPE MISMATCH\n";
			$output .= ">>>   tst mention = '$ref_mention->{MAP}->{head}{text}'";
			$output .= " ($ref_mention->{MAP}->{head}{start}:$ref_mention->{MAP}->{head}{end})";
			$output .= " ($ref_mention->{MAP}->{TYPE}) -- MENTION TYPE MISMATCH\n";
		    }
		    if (span_endpoint_mismatch ($ref_mention->{extent}, $ref_mention->{MAP}->{extent})
			> $max_mention_diff_chars) {
			$print_data ||= $print_err_data || $print_men_data;
			@pad = extent_alignment ($ref_mention->{extent}, $ref_mention->{MAP}{extent});
 			$output .= ">>>    ref mention extent = $pad[0]'$ref_mention->{extent}{text}'$pad[1]";
			$output .= " ($ref_mention->{extent}{start}:$ref_mention->{extent}{end}) -- MENTION EXTENT MISMATCH\n";
			$output .= ">>>    tst mention extent = $pad[2]'$ref_mention->{MAP}->{extent}{text}'$pad[3]";
			$output .= " ($ref_mention->{MAP}{extent}{start}:$ref_mention->{MAP}{extent}{end}) -- MENTION EXTENT MISMATCH\n";
		    }
		}
 		else {
		    $print_data ||= $print_err_data || $print_men_data;
		    $output .= ">>>   ref mention = '$ref_mention->{head}{text}'";
		    $output .= " ($ref_mention->{head}{start}:$ref_mention->{head}{end})";
		    $output .= " ($ref_mention->{TYPE}) -- NO MATCHING TST MENTION\n";
		}
	    }
	    foreach $tst_mention (@$tst_mentions) {
		if (not $tst_mention->{MAP}) {
		    $print_data ||= $print_err_data || $print_men_data;
		    $output .= ">>>   tst mention = '$tst_mention->{head}{text}'";
		    $output .= " ($tst_mention->{head}{start}:$tst_mention->{head}{end})";
		    $output .= " ($tst_mention->{TYPE}) -- NO MATCHING REF MENTION\n";
		}
	    }
	    $ref_names = $ref_entity->{names};
	    foreach $ref_name (@$ref_names) {
		if ($ref_name->{MAP}) {
		    if (span_endpoint_mismatch($ref_name, $ref_name->{MAP})
			<= $max_name_diff_chars) {
			$output .= "         ref name = '$ref_name->{text}'";
			$output .= " ($ref_name->{start}:$ref_name->{end})\n";
		    }
		    else {
			$print_data ||= $print_err_data || $print_nam_data;
			@pad = extent_alignment ($ref_name, $ref_name->{MAP});
			$output .= ">>>       ref name extent = $pad[0]'$ref_name->{text}'$pad[1]";
			$output .= " ($ref_name->{start}:$ref_name->{end}) -- NAME EXTENT MISMATCH\n";
			$output .= ">>>       tst name extent = $pad[2]'$ref_name->{MAP}{text}'$pad[3]";
			$output .= " ($ref_name->{MAP}{start}:$ref_name->{MAP}{end}) -- NAME EXTENT MISMATCH\n";
		    }
		}
		else {
		    $print_data ||= $print_err_data || $print_nam_data;
		    $output .= ">>>      ref name = '$ref_name->{text}'";
		    $output .= " ($ref_name->{start}:$ref_name->{end}) -- NO MATCHING TST NAME\n";
		}
	    }
	    $tst_names = $ref_entity->{MAP}{names};
	    foreach $tst_name (@$tst_names) {
		if (not $tst_name->{MAP}) {
		    $print_data ||= $print_err_data || $print_nam_data;
		    $output .= ">>>      tst name = '$tst_name->{text}'";
		    $output .= " ($tst_name->{start}:$tst_name->{end}) -- NO MATCHING REF NAME\n";
		}
	    }
	}
	else {
	    $print_data ||= $print_err_data || $print_ent_data;
	    $output .= ">>> ref entity $ref_entity->{ID} ($ref_entity->{TYPE}) -- NO MATCHING TST ENTITY\n";
	    $ref_mentions = $ref_entity->{mentions};
	    foreach $ref_mention (@$ref_mentions) {
		$output .= "      ref mention = '$ref_mention->{head}{text}'";
		$output .= " ($ref_mention->{head}{start}:$ref_mention->{head}{end})";
		$output .= " ($ref_mention->{TYPE}) -- unscored\n";
	    }
	}
	print $output if $print_data;
    }
    foreach $tst_entity (@$tst_entities) {
	$print_data = $print_all_data;
	if (not $tst_entity->{MAP}) {
	    $output = $separator;
	    $print_data ||= $print_all_data || $print_err_data || $print_ent_data;
	    $output .= ">>> tst entity $tst_entity->{ID} ($tst_entity->{TYPE}) -- NO MATCHING REF ENTITY\n";
	    $tst_mentions = $tst_entity->{mentions};
	    foreach $tst_mention (@$tst_mentions) {
		$output .= "      tst mention = '$tst_mention->{head}{text}'";
		$output .= " ($tst_mention->{head}{start}:$tst_mention->{head}{end})";
		$output .= " ($tst_mention->{TYPE}) -- unscored\n";
	    }
	    print $output if $print_data;
	}
    }
}

#################################

sub map_EDT_entities {

    my ($ref_entities, $tst_entities) = @_;
    my ($ref_entity, $tst_entity);
    my (@ref_cohorts, @tst_cohorts);
    my $null_candidate = {entity => {}, score => 0};

#map all reference entities
    foreach my $entity (@$ref_entities) {

#unless the entity has already been mapped
	unless (exists $entity->{MAP}) {
	    print "-" x 40, "\n" if $print_mapping_detail;
    
#determine the cohort set of possible mappings for this ref entity
	    tag_cohorts ($entity, $ref_entities, $tst_entities);
	    @ref_cohorts = collect_cohorts ($ref_entities);
	    @tst_cohorts = collect_cohorts ($tst_entities);

#order tst entities so that those with the best potential match come first
	    foreach $ref_entity (@ref_cohorts) {
		push my (@map_candidates), $null_candidate;
		foreach $tst_entity (@tst_cohorts) {
		    my $score = entity_mapping_score ($ref_entity, $tst_entity);
		    push @map_candidates, {entity => $tst_entity, score => $score}
		        if ($score > 0);
		}
		my @tst_ordered = sort {$b->{score} <=> $a->{score}} @map_candidates;
		$ref_entity->{tst_ordered} = [@tst_ordered];
		$ref_entity->{max_score} = $ref_entity->{tst_ordered}[0]->{score};
	    }
	    
#order ref entities so that those with the largest potential score come first
	    @ref_cohorts = sort {$b->{max_score} <=> $a->{max_score}} @ref_cohorts;
	    my $score_bound = 0;
	    foreach $ref_entity (reverse @ref_cohorts) {
		$ref_entity->{score_bound} = $score_bound;
		$score_bound += $ref_entity->{max_score};
	    }

#perform exhaustive search for optimum mapping over the cohort set
	    print_mapping_header (\@ref_cohorts, \@tst_cohorts)
		if $print_mapping_detail;
	    find_optimum_mapping ($best_score=0, \@ref_cohorts,
				  \@tst_cohorts, \@ref_cohorts);

#create test entity backpointers for all ref-tst mappings
	    foreach $ref_entity (@ref_cohorts) {
		if ($ref_entity->{MAP}) {
		    $ref_entity->{MAP}->{MAP} = $ref_entity;
		}
		else {
		    $ref_entity->{MAP} = 0;
		}
	    }
	}
    }
    foreach my $entity (@$ref_entities) {
	delete $entity->{MAP} unless $entity->{MAP};
    }
}

#################################

sub find_optimum_mapping {

    my ($prior_score, $ref_entities, $tst_entities, $ref_cohorts) = @_;
    my ($ref_entity, $tst_entity);
    my (@ref_remaining) = @$ref_entities;

#map the next ref entity
    if ($ref_entity = shift (@ref_remaining)) {
	return if ($prior_score+$ref_entity->{max_score} <= $best_score);
	
#now try all mappings
	my @tst_ordered = @{$ref_entity->{tst_ordered}};
	while (my $map_candidate = shift @tst_ordered) {

#each test entity may map to only one ref entity
	    $tst_entity = $map_candidate->{entity};
	    unless ($tst_entity->{MAP}) {
		
#try this mapping
		$tst_entity->{MAP} = $ref_entity if $tst_entity->{ID};
		find_optimum_mapping ($map_candidate->{score}+$prior_score,
				      \@ref_remaining, $tst_entities, $ref_cohorts);
		delete $tst_entity->{MAP};
	    }
	}
    }
	
#all reference entities have been mapped
    else {
	print_mapping_detail ($prior_score, $tst_entities)
	    if $print_mapping_detail;
	if ($prior_score > $best_score) { #keep better mapping
	    $best_score = $prior_score;
	    foreach my $entity (@$ref_cohorts) { #delete old mapping
		delete $entity->{MAP};
	    }
	    foreach my $entity (@$tst_entities) { #create new mapping
		$entity->{MAP}->{MAP} = $entity
		    if ($entity->{MAP});
	    }
	}
    }
}

#################################

sub print_mapping_header {

    my ($ref_cohorts, $tst_cohorts) = @_;

    foreach my $entity (@$ref_cohorts) {
	print "ref=$entity->{ID}, max_score=$entity->{max_score}, ".
	    "score_bound=$entity->{score_bound}\n";
	foreach my $map_candidate (@{$entity->{tst_ordered}}) {
	    print "   tst=$map_candidate->{entity}->{ID}, score=$map_candidate->{score}\n"
		if ($map_candidate->{score} > 0);
	}
    }
    print "REF cohort:";
    foreach my $entity (@$ref_cohorts) {
	print " ", substr($entity->{ID}, length($entity->{ID})-2, 2);
    }
    print "\nTST cohort:";
    foreach my $entity (@$tst_cohorts) {
	print " ", substr($entity->{ID}, length($entity->{ID})-2, 2);
    }
    print " score\n";
}

#################################

sub print_mapping_detail {

    my ($score, $tst_entities) = @_;

    print "       ref:";
    foreach my $entity (@$tst_entities) {
	$entity->{MAP} ?
	    print " ", substr($entity->{MAP}->{ID}, length($entity->{MAP}->{ID})-2, 2) :
		print "   ";
    }
    print "  $score",	($score > $best_score) ? "<=max\n" : "\n";
}

#################################

sub entity_mapping_score {

#N.B.  The entity mapping score must be 0 if tst doesn't match ref.

    my ($ref_entity, $tst_entity) = @_;
    my ($ref_mentions, $tst_mentions);
    my ($ref_mention, $tst_mention);

#check for null entity
    return 0 if not ($ref_entity and $tst_entity);

    my $score = $entity_mapping_score{$ref_entity}{$tst_entity};

#use table-lookup for precomputed scores
    return $score if defined $score;

#find the matching test mention for each reference mention
    $score = 0;
    $ref_mentions = $ref_entity->{mentions};
    $tst_mentions = $tst_entity->{mentions};
    foreach $ref_mention (@$ref_mentions) {
	foreach $tst_mention (@$tst_mentions) {
	    $score += mention_mapping_score 
		($ref_mention, $tst_mention);
	}
    }
    $entity_mapping_score{$ref_entity}{$tst_entity} = $score;
    return $score;
}
      
#################################

sub mention_mapping_score {

#N.B.  The mention mapping score must be 0 if tst doesn't match ref.

    my ($ref_mention, $tst_mention) = @_;
    my $mention_score;

#a test mention matches a ref mention if and only if:
#  the test mention points to the head of the ref mention
#--------
#The algorithm for determining whether the test mention "points to"
#the reference mention is to compute the mid-point of the test mention
#and to require this mid-point to lie within the span of the reference
#mention.

    $mention_score = 0;
    if (span_overlap($ref_mention->{head}, $tst_mention->{head})
	>= $min_head_overlap) {
	$mention_score = 
	    $ref_mention->{TYPE} eq "NAME" ? $nam_match :
		$ref_mention->{TYPE} eq "NOMINAL" ? $nom_match :
		    $ref_mention->{TYPE} eq "PRONOUN" ? $pro_match : 0;
    }
    return $mention_score;
}
      
#################################
      
sub collect_cohorts {

    my ($entities) = @_;
    my ($entity, @cohorts);

    foreach $entity (@$entities) {
	if (exists($entity->{cohort})) {
	    push @cohorts, $entity;
	    delete $entity->{cohort};
	}
    }
    return @cohorts;
}

#################################

sub tag_cohorts {

    my ($ent, $ent_set, $map_set) = @_;
    my $ent_map;

    $ent->{cohort} = "";
    foreach $ent_map (@$map_set) {
	unless (exists($ent_map->{cohort})) {
	    if (entity_mapping_score ($ent, $ent_map)) {
		tag_cohorts ($ent_map, $map_set, $ent_set);
	    }
	}
    }
}

#################################

sub span_overlap {

#This subroutine returns the minimum mutual overlap between two spans.

    my ($ref_span, $tst_span) = @_;
    my ($ref_start, $ref_end, $tst_start, $tst_end);
    my ($ref_overlap, $tst_overlap);
    
    $ref_start = $ref_span->{start};
    $ref_end = $ref_span->{end};
    $tst_start = $tst_span->{start};
    $tst_end = $tst_span->{end};

#compute overlap of ref_span by tst_span
    $tst_overlap = $tst_start > $ref_start ?
	$ref_end - $tst_start + 1 :
	    $tst_end - $ref_start + 1;
    $tst_overlap += $maa_chars;
    $tst_overlap /= ($ref_end - $ref_start + 1);

#compute overlap of tst_span by ref_span
    $ref_overlap = $ref_start > $tst_start ?
	$tst_end - $ref_start + 1 :
	    $ref_end - $tst_start + 1;
    $ref_overlap += $maa_chars;
    $ref_overlap /= ($tst_end - $tst_start + 1);

#return the minimum mutual overlap
    return min($tst_overlap, $ref_overlap);
}

#################################

sub span_endpoint_mismatch {

#This subroutine returns the maximum endpoint mismatch
#between two spans.

    my ($ref_span, $tst_span) = @_;
    my $max_diff = 0;

    $max_diff =
	max(abs($ref_span->{start} - $tst_span->{start}),
	    abs($ref_span->{end} - $tst_span->{end}));
}

#################################

sub max {

    my ($max, $x);

    $max = shift @_;
    foreach $x (@_) {
	if ($x > $max) {$max = $x}
    }
    return $max;
}

#################################

sub min {

    my ($min, $x);

    $min = shift @_;
    foreach $x (@_) {
	if ($x < $min) {$min = $x}
    }
    return $min;
}

#################################

sub get_EDT_ref_data {

    my ($ref_docs, $file) = @_;
    my ($line, $name);
    my ($data, $tag, $span, $type);
    my ($source, $doc_id, %document, @entities);
    my $ndocs = 0;

#read data from file
    open (FILE, $file) or die "\nUnable to open EDT output file '$input_file'", $usage;
    while ($line=<FILE>){
	$data .= $line;
    }
    close (FILE);

#get source file information
    ($tag, $span) = 
	extract_sgml_tag_and_span ("source_file", $data);
    die "\n\nFATAL INPUT ERROR:  no source file tag attribute '$name' in '$input_file'\n\n"
	unless (($type) = extract_sgml_tag_attribute ($name="TYPE", $tag));
    die "\n\nFATAL INPUT ERROR:  no source file tag attribute '$name' in '$input_file'\n\n"
	unless (($source) = extract_sgml_tag_attribute ($name="URI", $tag));

#get document data for all documents in the source file
    while (($tag, $span, $data) = 
	   extract_sgml_tag_and_span ("document", $data))
    {
	die "\n\nFATAL INPUT ERROR:  no document ID found in tag '$tag' in file '$input_file'\n\n"
	    unless (($doc_id) = extract_sgml_tag_attribute ("DOCID", $tag));
	die "\n\nFATAL INPUT ERROR:  document ID '$doc_id' in file '$input_file' is not unique\n\n"
	    if defined ($ref_docs->{$doc_id});
	$input_doc = $doc_id;
	$document{TYPE} = $type;
	$document{FILE} = $source;
	$document{entities} = [get_ref_entities ($span)];
	($ref_docs->{$doc_id}) = {%document};
	$ndocs++;
    }
    print "\n\nWARNING:  file '$input_file' contains no documents\n\n"
	if ($ndocs eq 0);
}

#################################

sub get_ref_entities {

#extract information for all entities in the document

    my (%entity, @entities);
    my ($tag, $data, $type);
    my $nentities = 0;

    while (($tag, $data, @_) =
	   extract_sgml_tag_and_span ("entity", @_))
    {
	die "\n\nFATAL INPUT ERROR:  no entity ID found for an entity in file '$input_file':\n\n".
	    "Entity tag is '$tag'\nEntity data is '$data'\n"
		unless (($input_entity) = extract_sgml_tag_attribute ("ID", $tag));
	die "\n\nFATAL INPUT ERROR for entity '$input_entity' in document '$input_doc'".
	    " in file '$input_file':\n  no entity TYPE found.  Entity data is '$data'\n\n"
		unless (($type) = extract_sgml_span ("entity_type", $data));

	$entity{ID} = $input_entity;
	$entity{TYPE} = $type eq "GSP" ? "GPE" : $type;
	$entity{mentions} = [get_ref_mentions ($data)];
	$entity{names} = [get_ref_names ($data)];

	push @entities, {%entity};
	$nentities++;
    }
    undef(%entity_mapping_score);
    print "\n\nWARNING:  document '$input_doc' in file '$input_file' contains no entities\n\n"
	if ($nentities eq 0);
    
    return @entities;
}

#################################

sub get_ref_mentions {

    my (%mention, @mentions);
    my ($tag, $data, $type);
    my ($start, $end);
    my $text;
    my $nmentions = 0;
    
    while (($tag, $data, @_) =
	   extract_sgml_tag_and_span ("entity_mention", @_))
    {
	die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	    "No mention TYPE found in tag '$tag'\n\n"
		unless (($type) = extract_sgml_tag_attribute ("TYPE", $tag));
	$mention{"TYPE"} = $type;
	
	($start, $end, $text) = extract_text_info ("head", $data);
	$mention{"head"} = {"start", $start, "end", $end, "text", $text};

	($start, $end, $text) = extract_text_info ("extent", $data);
	$mention{"extent"} = {"start", $start, "end", $end, "text", $text};

	push @mentions, {%mention};
	$nmentions++;
    }
    die "\n\nFATAL INPUT ERROR for entity '$input_entity' in document '$input_doc'".
	"in file '$input_file':\n  entity contains no mentions\n\n"
	    if ($nmentions eq 0);
    
    return @mentions;
}

#################################
    
sub get_ref_names {

    my $span;
    my (%name, @names);
    my ($tag, $data, $start, $end, $text);
    
    ($span) = extract_sgml_span ("entity_attributes", @_);

    while ($span and (($tag, $data, $span) = 
	   extract_sgml_tag_and_span ("name", $span)))
    {
	($start, $end, $text) = extract_text_info ("charseq", $data);
	%name = ("start", $start, "end", $end, "text", $text);

	push @names, {%name};
    }
    return @names;
}

#################################
    
sub extract_text_info {

    my ($tag, $span) = @_;
    my ($data, $start, $end, $text);

    die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	"No tag '$tag' found in span '$span'\n\n"
	    unless (($data) = extract_sgml_span ($tag, $span));
    die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	"No tag start found in data '$data'\n\n"
	    unless (($start) = extract_sgml_span ("start", $data));
    die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	"No tag end found in data '$data'\n\n"
	    unless (($end) = extract_sgml_span ("end", $data));
    ($text) = extract_sgml_tag_attribute ("string", $data);
    return ($start, $end, $text);
}

#################################

sub extract_sgml_span {

    my ($tag_name, $input) = @_;
    my $span;

    unless ($input =~ m|<$tag_name\s*.*?>\s*|s) {return;}
    $span = $';
    unless ($span =~ m|\s*</$tag_name>\s*|s) {return;}
    $span = $`;
    return ($span, $');
}

#################################

sub extract_sgml_tag {

    my ($tag_name, $input) = @_;
    my $tag;

    unless ($input =~ m|<$tag_name\s*|s) {return;}
    $tag = $';
    unless ($tag =~ m|\s*>\s*|) {return;}
    $tag = $`;
}

#################################

sub extract_sgml_tag_and_span {
    
    my ($tag_name, $input) = @_;
    my ($tag, $span);
    
    unless ($input =~ m|<$tag_name\s*|s) {return;}
    $tag = $';
    unless ($tag =~ m|\s*>|) {return;}
    $tag = $`;
    $span = $';
    if ($span =~ m|\s*</$tag_name>\s*|s) {
	$span = $`;
	($tag, $span, $');
    }
    else {
	($tag);
    }
}

#################################

sub extract_sgml_tag_attribute {

    my ($attribute, $tag_name) = @_;
    my $value;

    unless ($tag_name =~ m|$attribute[\s=]+\"|) {return;}
    $value = $';
    unless ($value =~ m|\"|) {return;}
    $value = $`;
}

#################################

sub date_time_stamp {

    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime();
    my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
    my ($date, $time);

    $time = sprintf "%2.2d:%2.2d:%2.2d", $hour, $min, $sec;
    $date = sprintf "%4.4s %3.3s %s", 1900+$year, $months[$mon], $mday;
    return ($date, $time);
}

#################################

sub extent_alignment {

    my ($extent1, $extent2) = @_;

    return (" " x ($extent1->{start} - $extent2->{start}),
	    " " x ($extent2->{end}   - $extent1->{end}),
	    " " x ($extent2->{start} - $extent1->{start}),
	    " " x ($extent1->{end}   - $extent2->{end}));
}

#################################

sub print_EDT_ref_data {

    my $ref_docs = @_;
    my (%ref_docs, $doc_id, %document, $document);
    my ($entity, $mention, $name);
    
    while (@_) {
	$doc_id = shift(@_);
	$document = shift(@_);
	print "doc ID = '$doc_id', ";
	print "TYPE = '$document->{TYPE}', ";
	print "FILE = '$document->{FILE}'\n";
	foreach $entity (@{$document->{entities}}) {
	    print "  entity ID = '$entity->{ID}', ";
	    print "TYPE = '$entity->{TYPE}'\n";
	    foreach $mention (@{$entity->{mentions}}) {
		print "    mention TYPE = '$mention->{TYPE}', ";
		print "head = $mention->{head}{start}:$mention->{head}{end}",
		    " ($mention->{head}{text}), ";
		print "extent = $mention->{extent}{start}:$mention->{extent}{end}",
		    " ($mention->{extent}{text})\n";
	    }
	    foreach $name (@{$entity->{names}}) {
		print "      name extent = $name->{start}:$name->{end}",
		    " ($name->{text})\n";
	    }
	}
    }
}

#################################

sub output_entity_pairs {

    my ($out_file, $ref) = @_;
    my ($doc_id, $ref_entity, $ref_entities);

    open (OUTPUT, ">$out_file") or die
	"\nUnable to open ref-tst entity pair output file '$out_file'", $usage;
    foreach $doc_id (keys %$ref) {
	$ref_entities = $ref->{$doc_id}{entities};
	foreach $ref_entity (@$ref_entities) {
	    if ($ref_entity->{MAP}) {
		print OUTPUT "\"$ref_entity->{ID}\" \"$ref_entity->{MAP}{ID}\"\n";
	    }
	}
    }
    close (OUTPUT);
}

#################################

sub get_EDT_source_data {

    my ($ref_docs, $file) = @_;
    my ($line, $data, $tag);
    my ($source, $doc_id);
    my $gray = "#999999";

#read data from file
    open (FILE, $file) or die "\nUnable to open EDT source data file '$input_file'", $usage;
    while ($line=<FILE>){
	$data .= $line;
    }
    close (FILE);

#make sure line termination is just "\n".
    $data =~ s/[\013]([\n]?)/\n/g;

#create character pointers, skipping sgml markup
    my @source = split //, $data; 
    my @pointer;
    my $markup = 0;
    my $count = 0;
    for (my $i=0; $i<@source; $i++) {
	$markup ||= $source[$i] eq "<";
	$pointer[$count++] = $i unless $markup;
	$markup = 0 if $source[$i] eq ">";
	$source[$i] = "<i><font color=$gray>&lt;" if $source[$i] eq "<";
	$source[$i] = "></font></i>" if $source[$i] eq ">";
	$source[$i] = "</tt><br><tt>" if $source[$i] eq "\n";
    }

#save source text for each document
    while (($tag, $doc_id, $data) =
	   extract_sgml_tag_and_span ("DOCNO", $data)) {
	$doc_id =~ s/ *$//;
	$doc_id =~ s/^ *//;
	if (defined $ref_docs->{$doc_id}) {
	    $ref_docs->{$doc_id}{TEXT} = [@source];
	    $ref_docs->{$doc_id}{PNTR} = [@pointer];
	}
    }
}

#################################

sub output_entity_comparisons {

    my ($out_dir, $ref_docs, $tst_docs) = @_;
    my $doc_id;
    my ($ref_doc, $tst_doc);
    my ($ref_entities, $tst_entities);
    my ($ref_entity, $tst_entity);
    my ($ref_cohort, $tst_cohort);
    my ($ref_mention, $tst_mention, $ref_name, $tst_name);
    my ($i, $j, $char);
    my @colors = ("\"#FF0000\"", "\"#3333FF\"", "\"#009900\"", "\"#CC33CC\"", "\"#CC6600\"",
		  "\"#999900\"", "\"#33CCFF\"", "\"#FFCC33\"", "\"#FF99FF\"", "\"#66FF99\"");
    my $color;
    my (@color_ref, @color_tst);
    my (@type_ref, @type_tst);
    my $space = "&nbsp;";
    my $space4 = $space x 4;
    my $space6 = $space x 6;
    my $tst_id;
    my $err_on = "<span class=bg><u>";
    my $err_off = "</u></span>";

    my $header1 = "<html><head><title>";
#     $title
    my $header2 = "</title></head><style type=\"text/css\"><!--.bg {background-color: yellow}--></style>".
	"<body><blockquote><center><font size=+1>$err_on underline => Ref and Tst differ".
	"$err_off".$space6."color => extent".$space6." <b>+bold => head".$space6."<i>+italics => name</i></b>".
	    "</font></center>";
    my $hdr_tbl = "<center><table BORDER COLS=2 WIDTH=\"95%\" ><tr><td ALIGN=LEFT VALIGN=TOP>".
		"<font color=\"#000000\"><font size=+1>Color coding for mentions of REF entities:".
		    "</font></font>";
    my $hdr_color = "<br><b><font color=";
#     $color
    my $hdr_ent = "><font size=+1>".$space4;
#     $entity
    my $tlr_ent = "</font></font></b>";
    my $header3 = "</td><td ALIGN=LEFT VALIGN=TOP><font color=\"#000000\"><font size=+1>Color coding ".
	"for mentions of TST entities:</font></font>";
#     $tst_colors/entities
    my $hdr_ref = "</td></tr><tr><td ALIGN=LEFT VALIGN=TOP><tt>";
#     $ref_chars
    my $hdr_tst = "</tt></td><td ALIGN=LEFT VALIGN=TOP><tt>";
#     $tst_chars
    my $tlr_tbl = "</tt></td></tr></table></center>";
    my $tlr_doc = "</blockquote></body></html>";
    my $nil= $space4."(nil)";
#compute output for all ref entities
    foreach $doc_id (keys %$ref_docs) {
	my $first = 1;
	$ref_doc = $ref_docs->{$doc_id};
	$ref_entities = $ref_doc->{entities};
	$tst_doc = $tst_docs->{$doc_id};
	$tst_entities = $tst_doc->{entities};
	my @pntr = @{$ref_doc->{PNTR}};
	my (%ref_done, %tst_done);
	foreach $ref_entity (@{$ref_doc->{entities}}) {
	    next if $ref_done{$ref_entity->{ID}};
	    tag_cohorts ($ref_entity, $ref_entities, $tst_entities);
	    my @ref_cohorts = collect_cohorts ($ref_entities);
	    my @tst_cohorts = collect_cohorts ($tst_entities);

	    my $count = 0;
	    my (%ref_color, %tst_color);
	    my @chr_type = split //, 0 x @{$ref_doc->{TEXT}};
	    my @chr_diff = @chr_type;
	    my @chr_color = @chr_type;
	    my $errs_exist = 0;

#find differences between ref and tst annotation for all ref cohorts
	    foreach $ref_cohort (@ref_cohorts) {
		$color = $ref_color{$ref_cohort->{ID}} = $colors[$count];
		my $mapped = defined $ref_cohort->{MAP};
		$errs_exist ||= not $mapped;
		if ($mapped) {
		    if ($ref_cohort->{MAP}{TYPE} eq $ref_cohort->{TYPE}) {
			$color_ref[$count] = $ref_cohort->{TYPE}.$space4.$ref_cohort->{ID};
			$color_tst[$count] = $ref_cohort->{TYPE}.$space4.$ref_cohort->{MAP}{ID};
		    }
		    else {
			$color_ref[$count] = $err_on.$ref_cohort->{TYPE}.$err_off.$space4.$ref_cohort->{ID};
			$color_tst[$count] = $err_on.$ref_cohort->{MAP}{TYPE}.$err_off.$space4.$ref_cohort->{MAP}{ID};
		    }
		}
		else {
		    $color_ref[$count] = $ref_cohort->{TYPE}.$space4.$ref_cohort->{ID};
		    $color_tst[$count] = $nil;
		}
		$count++;
		$ref_done{$ref_cohort->{ID}} = 1;
		foreach $ref_mention (@{$ref_cohort->{mentions}}) {
		    $tst_mention = $ref_mention->{MAP};
		    for ($i=$ref_mention->{extent}{start}; $i<=$ref_mention->{extent}{end}; $i++) {
			if ($chr_type[$pntr[$i]] eq 0) {
			    $chr_type[$pntr[$i]] = "extent" if $chr_type[$pntr[$i]] eq 0;
			    $chr_color[$pntr[$i]] = $color;
			}
			$chr_diff[$pntr[$i]] ||= (not $mapped) || (not defined $tst_mention) ||
			    $i < $tst_mention->{extent}{start} || $i > $tst_mention->{extent}{end};
		    }
		    for ($i=$ref_mention->{head}{start}; $i<=$ref_mention->{head}{end}; $i++) {
			if ($chr_type[$pntr[$i]] ne "name") {
			    $chr_type[$pntr[$i]] = "head";
			    $chr_color[$pntr[$i]] = $color;
			}
			$chr_diff[$pntr[$i]] ||= (not $mapped) || (not defined $tst_mention) ||
			    $i < $tst_mention->{head}{start} || $i > $tst_mention->{head}{end};
		    }
		}
		foreach $ref_name (@{$ref_cohort->{names}}) {
		    $tst_name = $ref_name->{MAP};
		    for ($i=$ref_name->{start}; $i<=$ref_name->{end}; $i++) {
			$chr_type[$pntr[$i]] = "name";
			$chr_color[$pntr[$i]] = $color;
			$chr_diff[$pntr[$i]] ||= (not $mapped) || (not defined $tst_name) ||
			    $i < $tst_name->{start} || $i > $tst_name->{end};
		    }
		}
	    }

#output ref annotation with differences between ref and tst highlighted
	    my @ref_chars = @{$ref_doc->{TEXT}};
	    for ($j=0; $j<@{$ref_doc->{TEXT}}; $j++) {
		my $char = $ref_chars[$j];
		$errs_exist ||= $chr_diff[$j];
		$char = "$err_on$char$err_off" if $chr_diff[$j];
		$char = "<font color=$chr_color[$j]>$char</font>" if $chr_type[$j];
		$char = $chr_type[$j] eq "name" ? "<i><b>$char</b></i>" :
		    $chr_type[$j] eq "head" ? "<b>$char</b>" : $char;
		$ref_chars[$j] = $char;
	    }
		
#find differences between ref and tst annotation for all tst cohorts
	    @chr_type = split //, 0 x @{$ref_doc->{TEXT}};
	    @chr_diff = @chr_type;
	    @chr_color = @chr_type;
	    foreach $tst_cohort (@tst_cohorts) {
		my $mapped = defined $tst_cohort->{MAP};
		$errs_exist ||= not $mapped;
		if ($mapped) {
		    $color = $tst_color{$tst_cohort->{ID}} = 
			$ref_color{$tst_cohort->{MAP}{ID}};
		}
		else{
		    $color = $tst_color{$tst_cohort->{ID}} = $colors[$count];
		    $color_ref[$count] = $nil;
		    $color_tst[$count] = $tst_cohort->{TYPE}.$space4.$tst_cohort->{ID};
		    $count++;
		}
		$tst_done{$tst_cohort->{ID}} = 1;
		foreach $ref_mention (@{$tst_cohort->{mentions}}) {
		    $tst_mention = $ref_mention->{MAP};
		    for ($i=$ref_mention->{extent}{start}; $i<=$ref_mention->{extent}{end}; $i++) {
			if ($chr_type[$pntr[$i]] eq 0) {
			    $chr_type[$pntr[$i]] = "extent" if $chr_type[$pntr[$i]] eq 0;
			    $chr_color[$pntr[$i]] = $color;
			}
			$chr_diff[$pntr[$i]] ||= (not $mapped) || (not defined $tst_mention) ||
			    $i < $tst_mention->{extent}{start} || $i > $tst_mention->{extent}{end};
		    }
		    for ($i=$ref_mention->{head}{start}; $i<=$ref_mention->{head}{end}; $i++) {
			if ($chr_type[$pntr[$i]] ne "name") {
			    $chr_type[$pntr[$i]] = "head";
			    $chr_color[$pntr[$i]] = $color;
			}
			$chr_diff[$pntr[$i]] ||= (not $mapped) || (not defined $tst_mention) ||
			    $i < $tst_mention->{head}{start} || $i > $tst_mention->{head}{end};
		    }
		}
		foreach $ref_name (@{$tst_cohort->{names}}) {
		    $tst_name = $ref_name->{MAP};
		    for ($i=$ref_name->{start}; $i<=$ref_name->{end}; $i++) {
			$chr_type[$pntr[$i]] = "name";
			$chr_color[$pntr[$i]] = $color;
			$chr_diff[$pntr[$i]] ||= (not $mapped) || (not defined $tst_name) ||
			    $i < $tst_name->{start} || $i > $tst_name->{end};
		    }
		}
	    }

#output tst annotation with differences between ref and tst highlighted
	    my @tst_chars;
	    if (@tst_cohorts) {
		@tst_chars = @{$ref_doc->{TEXT}};
		for ($j=0; $j<@{$ref_doc->{TEXT}}; $j++) {
		    my $char = $tst_chars[$j];
		    $errs_exist ||= $chr_diff[$j];
		    $char = "$err_on$char$err_off" if $chr_diff[$j];
		    $char = "<font color=$chr_color[$j]>$char</font>" if $chr_type[$j];
		    $char = $chr_type[$j] eq "name" ? "<i><b>$char</b></i>" :
			$chr_type[$j] eq "head" ? "<b>$char</b>" : $char;
		    $tst_chars[$j] = $char;
		}
	    }

#create output file
	    next if not ($print_all_data || ($errs_exist && $print_err_data));
	    if ($first) {
		my $x = chop $out_dir;
		$out_dir .= $x unless $x eq "/";
		my $file = "$out_dir/EDT_COMP_$doc_id.html";
		open (FILE, ">$file") or die "\nUnable to open html diagnostic output file '$file'";
		print FILE $header1.$doc_id.$header2;
		$first = 0;
	    }
	    print FILE $hdr_tbl;
	    for ($i=0; $i<$count; $i++) {
		print FILE $hdr_color, $colors[$i], $hdr_ent, $color_ref[$i], $tlr_ent;
	    }
	    print FILE $header3;
	    for ($i=0; $i<$count; $i++) {
		print FILE $hdr_color, $colors[$i], $hdr_ent, $color_tst[$i], $tlr_ent;
	    }
	    print FILE $hdr_ref, @ref_chars, $hdr_tst, @tst_chars, $tlr_tbl;
	}
	    
#compute output for tst entities that haven't been processed (i.e., with no matching ref entities)
	$tst_doc = $tst_docs->{$doc_id};
	foreach $tst_entity (@{$tst_doc->{entities}}) {
	    my $errs_exist = 1;
	    next if $tst_done{$tst_entity->{ID}};
	    tag_cohorts ($tst_entity, $tst_entities, $ref_entities);
	    my @tst_cohorts = collect_cohorts ($tst_entities);
	    my $count = 0;
	    my (%ref_color, %tst_color);
	    my @chr_type = split //, 0 x @{$ref_doc->{TEXT}};
	    my @chr_color = @chr_type;

#compute output for tst cohorts with no matching ref entities
	    foreach $tst_cohort (@tst_cohorts) {
		$color = $tst_color{$tst_cohort->{ID}} = $colors[$count];
		$color_ref[$count] = $nil;
		$color_tst[$count] = $tst_cohort->{TYPE}.$space4.$tst_cohort->{ID};
		$count++;
		$tst_done{$tst_cohort->{ID}} = 1;
		foreach $ref_mention (@{$tst_cohort->{mentions}}) {
		    for ($i=$ref_mention->{extent}{start}; $i<=$ref_mention->{extent}{end}; $i++) {
			if ($chr_type[$pntr[$i]] eq 0) {
			    $chr_type[$pntr[$i]] = "extent" if $chr_type[$pntr[$i]] eq 0;
			    $chr_color[$pntr[$i]] = $color;
			}
		    }
		    for ($i=$ref_mention->{head}{start}; $i<=$ref_mention->{head}{end}; $i++) {
			if ($chr_type[$pntr[$i]] ne "name") {
			    $chr_type[$pntr[$i]] = "head";
			    $chr_color[$pntr[$i]] = $color;
			}
		    }
		}
		foreach $ref_name (@{$tst_cohort->{names}}) {
		    for ($i=$ref_name->{start}; $i<=$ref_name->{end}; $i++) {
			$chr_type[$pntr[$i]] = "name";
			$chr_color[$pntr[$i]] = $color;
		    }
		}
	    }
		
#output tst annotation
	    my @tst_chars = @{$ref_doc->{TEXT}};
	    for ($j=0; $j<@{$ref_doc->{TEXT}}; $j++) {
		my $char = $tst_chars[$j];
		$char = "<font color=$chr_color[$j]>$err_on$char$err_off</font>" if $chr_type[$j];
		$char = $chr_type[$j] eq "name" ? "<i><b>$char</b></i>" :
		    $chr_type[$j] eq "head" ? "<b>$char</b>" : $char;
		$tst_chars[$j] = $char;
	    }

#create output file
	    next if not ($print_all_data || ($errs_exist && $print_err_data));
	    if ($first) {
		my $x = chop $out_dir;
		$out_dir .= $x unless $x eq "/";
		my $file = "$out_dir/EDT_COMP_$doc_id.html";
		open (FILE, ">$file") or die "\nUnable to open html diagnostic output file '$file'";
		print FILE $header1.$doc_id.$header2;
		$first = 0;
	    }
	    print FILE $hdr_tbl;
	    for ($i=0; $i<$count; $i++) {
		print FILE $hdr_color, $colors[$i], $hdr_ent, $color_ref[$i], $tlr_ent;
	    }
	    print FILE $header3;
	    for ($i=0; $i<$count; $i++) {
		print FILE $hdr_color, $colors[$i], $hdr_ent, $color_tst[$i], $tlr_ent;
	    }
	    print FILE $hdr_ref, $hdr_tst, @tst_chars, $tlr_tbl;
	}
	if (not $first) {
	    print FILE $tlr_doc;
	    close FILE;
	}
    }
}