#!/usr/bin/perl -w

use strict;

#SCORING PARAMETERS:=================================================

# min_overlap is the minimum mutual fractional overlap allowed
# for a mention head or a name to be declared as successfully detected.
use vars qw ($min_overlap);
    $min_overlap = 0.3; #minimum fractional overlap for mention detection

# max_diff is the maximum extent difference allowed for names and 
#mentions to be declared a "match".
use vars qw ($max_diff_chars $max_diff_time $max_diff_xy);
#       max_diff_chars is the maximum extent difference
#       in characters for text sources.
    $max_diff_chars = 4;
#       max_diff_time is the maximum extent difference
#       in seconds for audio sources.
    $max_diff_time = 0.4;
#       max_diff_xy is the maximum extent difference
#       in centimeters for image mentions.
    $max_diff_xy = 0.4;
# NB -- max_diff_xy is currently in pixels rather than cm because
#       the bounding box data are expressed in pixels rather than cm

#The following three parameters are used to map ref entities onto tst entities
# nam_match is the weight assigned to mentions of type NAME
# nom_match is the weight assigned to mentions of type NOMINAL
# pro_match is the weight assigned to mentions of type PRONOUN
use vars qw ($nam_match $nom_match $pro_match);
    $nam_match = 1.00;
    $nom_match = 0.20;
    $pro_match = 0.04;

#====================================================================

#################################
# History:
#
# version 00
#    This initial version is an adaptation of EDT_ref_compare.v23.pl
#    to support evaluation of EDT system output.  This evaluation
#    includes support for audio (ASR) and image (OCR) sources in
#    addition to text sources.  Support for visualizing differences
#    between EDT ref annotation and EDT output has been removed.
#
# version 01
#    This version modifies the representation of mention and name locators
#    to facilitate handling multimodal (e.g., char string, time and spatial)
#    locators.
#
# version 02
#    This version adds in the image locator structure to accommodate image
#    source data in which each image locator is a set of bounding boxes.
#    Mapping parameters are made a function of source type.
#
# version 03
#    This version processes all of the system EDT output documents rather
#    than all of the reference documents.  Thus it is only necessary that
#    the reference documents include the system EDT output documents under
#    evaluation.  They need not be identical.
#    This version also fixed errors in time mapping for audio and refined
#    the span overlap computation by including a nominal_frame_period and
#    nominal_pixel_size for audio and image source_file types, respectively.
#
# version 04
#    bug fix to correct the count of the number of mentions with correctly
#    recognized extents.
#
#################################

use vars qw ($print_entities $print_all_data $print_err_data);
use vars qw ($print_ent_data $print_men_data $print_nam_data);

use vars qw ($print_mapping_detail $print_timing);

use vars qw ($input_file $input_doc $input_entity);
use vars qw ($document);
use vars qw ($usage);
use vars qw ($best_score);
use vars qw (%entity_detection_statistics);
use vars qw (%entity_type_statistics);
use vars qw (%mention_detection_statistics);
use vars qw (%name_detection_statistics);
use vars qw (%entity_mapping_score);

use vars qw (@error_types @entity_types @mention_types @count_types);
@error_types = ("correct", "miss", "fa", "error");
@entity_types = ("FACILITY", "GPE", "LOCATION", "ORGANIZATION", "PERSON", "total");
@mention_types = ("NAME", "NOMINAL", "PRONOUN", "total");
@count_types = ("1", "2", "3-4", "5-8", ">8", "total");
#
#N.B.  The above values of entity_types, mention_types and count_types MUST BE ALL UNIQUE!!!
#      (This is because they are all keys in the SAME hash table.)
#

my $version = "04";

my ($date, $time) = date_time_stamp();
print "EDT evaluation (version $version) run on $date at $time\n";
print "command line:  ", $0, " ", join(" ", @ARGV), "\n";
printf "PARAMETERS:\n".
    "  min mutual fractional overlap of matching mention heads or names:\n".
    "%11.3f\n".
    "  entity mapping weights:\n".
    "%11.3f for NAME mentions\n".
    "%11.3f for NOMINAL mentions\n".
    "%11.3f for PRONOUN mentions\n".
    "  max allowable extent difference for names and mentions to match:\n".
    "%11d chars for text sources\n".
    "%11.3f sec for audio sources\n".
    "%11.3f cm for image sources\n",
    $min_overlap, $nam_match, $nom_match, $pro_match,
    $max_diff_chars, $max_diff_time, $max_diff_xy;

$usage = "\n\n$0 [-hsaxemn] -r <ref_list> -t <tst_list>\n\n".
    "Description:  This Perl program evaluates EDT system performance.\n".
    "\n".
    "Required arguments:\n".
    "  -r <ref_list> is a file containing a list of EDT reference data for\n".
    "      all documents to be evaluated.  These data must be in standard\n".
    "      EDT apf format\n".
    "  -t <tst_list> is a file containing a list of system EDT output data for\n".
    "      all documents to be evaluated.  These data must be in standard\n".
    "      EDT apf format\n".
    "\n".
    "Optional arguments:\n".
    "  -h prints this help message to STDOUT \n".
    "  -s prints EDT annotation data summary to STDOUT\n".
    "\n".
    "      The following flags control the output of comparison data for diagnosis:\n".
    "      There are two comparison outputs, a printed \"report\" of ref/tst differences\n".
    "      sent to STDOUT and an html \"illustration\" of entity mentions written to\n".
    "      directory <cmp_dir>, one file per document.  The html file name is\n".
    "      \"EDT_COMP_<doc_id>.html\"\n".
    "  -a outputs report data for all entities\n".
    "  -x outputs report data for entities with errors\n".
    "  -e outputs report data for entities with entity detection/type errors\n".
    "  -m outputs report data for entities with mention detection/extent errors\n".
    "  -n outputs report data for entities with name detection/extent errors\n".
    "\n".
    "      The following flags provide diagnostic output for program development:\n".
    "  -i outputs the time spent mapping ref entities to tst entities for each doc\n".
    "  -d outputs detailed mapping info and score for every mapping computed\n".
    "\n";

{
    use vars qw ($opt_h $opt_s $opt_a $opt_x $opt_e $opt_m $opt_n);
    use vars qw ($opt_r $opt_t $opt_i $opt_d);
    require "getopts.pl";
    &Getopts ('hsaxemnidr:t:');
    die $usage if defined($opt_h);
    $print_entities = defined($opt_s) ? 1 : 0;
    $print_all_data = defined($opt_a) ? 1 : 0;
    $print_err_data = defined($opt_x) ? 1 : 0;
    $print_ent_data = defined($opt_e) ? 1 : 0;
    $print_men_data = defined($opt_m) ? 1 : 0;
    $print_nam_data = defined($opt_n) ? 1 : 0;
    $print_timing = defined($opt_i) ? 1 : 0;
    $print_mapping_detail = defined($opt_d) ? 1 : 0;
    die "Error in command line$usage"
	unless (defined($opt_r) and defined($opt_t));

    my ($list0) = $opt_r;
    my ($list1) = $opt_t;
    my (%ref_docs, %tst_docs);

#read in all EDT reference data
    open (LIST, $list0) or die "\nUnable to open file list '$list0'", $usage;
    while ($input_file = <LIST>) {
	get_EDT_data (\%ref_docs, $input_file);
    }
    close (LIST);
    
#read in all EDT system output data
    open (LIST, $list1) or die "\nUnable to open file list '$list1'", $usage;
    while ($input_file = <LIST>) {
	get_EDT_data (\%tst_docs, $input_file);
    }
    close (LIST);
    
#print entity data
    if ($print_entities) {
	print "REF ", "=" x 30, " REF ", "=" x 30, " REF\n";
	print_EDT_data (%ref_docs);
	print "END ", "=" x 30, " END ", "=" x 30, " END\n";
	print "TST ", "=" x 30, " TST ", "=" x 30, " TST\n";
	print_EDT_data (%tst_docs);
	print "END ", "=" x 30, " END ", "=" x 30, " END\n";
    }
    
#compare input data and evaluate differences
    eval_EDT (\%ref_docs, \%tst_docs);
}    
    
#################################

sub eval_EDT {

    my ($ref, $tst) = @_;
    my $doc_id;
    my ($ref_doc, $tst_doc);
    my ($ref_entities, $tst_entities);

    initialize_statistics ();

    foreach $doc_id (keys %$tst) {

	$document = $tst_doc = $tst->{$doc_id};
	$ref_doc = $ref->{$doc_id};
	die "\n\nFATAL ERROR:  tst doc '$doc_id' has no corresponding reference doc\n\n"
	    unless  defined($ref_doc);
	die "\n\nFATAL ERROR:  ref and tst documents must be of the same type\n".
	    "For document '$doc_id', ref type = '$ref_doc->{TYPE}' but tst type = '$tst_doc->{TYPE}'\n\n",
	    unless $tst_doc->{TYPE} eq $ref_doc->{TYPE};

#map the entities and mentions
	$ref_entities = $ref_doc->{entities};
	$tst_entities = $tst_doc->{entities};
	undef(%entity_mapping_score); #N.B. -- this assumes no cross-doc entities
	my $tstart = (times)[0];
	map_EDT_entities ($ref_entities, $tst_entities);
	if ($print_timing) {
	    my $mapping_time = (times)[0] - $tstart;
	    my $nr = @$ref_entities;
	    my $nrm = 0;
	    foreach my $entity (@$ref_entities) {
		$nrm += @{$entity->{mentions}};
	    }
	    my $nt = @$tst_entities;
	    my $ntm = 0;
	    foreach my $entity (@$tst_entities) {
		$ntm += @{$entity->{mentions}};
	    }
	    printf "%10.2f secs to map%3d ref entities (%3d mentions) to%3d tst entities (%3d mentions) for doc %s\n",
	        $mapping_time, $nr, $nrm, $nt, $ntm, $doc_id;
	}

#tabulate scoring statistics
	tabulate_EDT_scoring_statistics ($ref_entities, $tst_entities);
	print_scoring_details ($ref_entities, $tst_entities)
	    if ($print_all_data or $print_err_data or $print_ent_data
		or $print_men_data or $print_nam_data);
    }

#score entity detection
    score_EDT_entity_detection ();
    
#score attribute recognition
    score_EDT_attribute_recognition ();

#score mention recognition
    score_EDT_mention_recognition ();
}

#################################

sub score_EDT_entity_detection {

    my $type;

#scoring conditioned on entity type
    foreach $type (@entity_types) {
	next if $type eq "total";
	$entity_detection_statistics{total}{miss} += $entity_detection_statistics{$type}{miss};
	$entity_detection_statistics{total}{fa} += $entity_detection_statistics{$type}{fa};
	$entity_detection_statistics{total}{correct} += $entity_detection_statistics{$type}{correct};
	$entity_detection_statistics{total}{error} += $entity_detection_statistics{$type}{error};
    }

    print "\nEntity Detection and Type Recognition statistics:\n",
    " ref     ____________count______________       ____________percent____________\n",
    " entity  Detection      Type_Recognition       Detection      Type_Recognition\n",
    " type    miss    fa     miss   err  corr       miss    fa     miss   err  corr\n";
    foreach $type (@entity_types) {
	my $total = ($entity_detection_statistics{$type}{miss} +
		     $entity_detection_statistics{$type}{error} +
		     $entity_detection_statistics{$type}{correct});
	printf "  %3.3s%8d%6d%9d%6d%6d%11.1f%6.1f%9.1f%6.1f%6.1f\n", $type,
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{fa},
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{error},
	$entity_detection_statistics{$type}{correct},
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{fa}/max($total,1),
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{error}/max($total,1),
	100*$entity_detection_statistics{$type}{correct}/max($total,1);
    }

#scoring conditioned on mention type
    $entity_detection_statistics{total}{miss} =
	$entity_detection_statistics{total}{fa} =
	    $entity_detection_statistics{total}{correct} =
		$entity_detection_statistics{total}{error} = 0;
	
    foreach $type (@mention_types) {
	next if $type eq "total";
	$entity_detection_statistics{total}{miss} += $entity_detection_statistics{$type}{miss};
	$entity_detection_statistics{total}{fa} += $entity_detection_statistics{$type}{fa};
	$entity_detection_statistics{total}{correct} += $entity_detection_statistics{$type}{correct};
	$entity_detection_statistics{total}{error} += $entity_detection_statistics{$type}{error};
    }

    print "\nEntity Detection and Type Recognition statistics:\n",
    " ref     ____________count______________       ____________percent____________\n",
    " mention Detection      Type_Recognition       Detection      Type_Recognition\n",
    " type    miss    fa     miss   err  corr       miss    fa     miss   err  corr\n";
    foreach $type (@mention_types) {
	my $total = ($entity_detection_statistics{$type}{miss} +
		     $entity_detection_statistics{$type}{error} +
		     $entity_detection_statistics{$type}{correct});
	printf "  %3.3s%8d%6d%9d%6d%6d%11.1f%6.1f%9.1f%6.1f%6.1f\n", $type,
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{fa},
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{error},
	$entity_detection_statistics{$type}{correct},
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{fa}/max($total,1),
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{error}/max($total,1),
	100*$entity_detection_statistics{$type}{correct}/max($total,1);
    }

#scoring conditioned on count type
    $entity_detection_statistics{total}{miss} =
	$entity_detection_statistics{total}{fa} =
	    $entity_detection_statistics{total}{correct} =
		$entity_detection_statistics{total}{error} = 0;
	
    foreach $type (@count_types) {
	next if $type eq "total";
	$entity_detection_statistics{total}{miss} += $entity_detection_statistics{$type}{miss};
	$entity_detection_statistics{total}{fa} += $entity_detection_statistics{$type}{fa};
	$entity_detection_statistics{total}{correct} += $entity_detection_statistics{$type}{correct};
	$entity_detection_statistics{total}{error} += $entity_detection_statistics{$type}{error};
    }

    print "\nEntity Detection and Type Recognition statistics:\n",
    " ref     ____________count______________       ____________percent____________\n",
    " mention Detection      Type_Recognition       Detection      Type_Recognition\n",
    " count   miss    fa     miss   err  corr       miss    fa     miss   err  corr\n";
    foreach $type (@count_types) {
	my $total = ($entity_detection_statistics{$type}{miss} +
		     $entity_detection_statistics{$type}{error} +
		     $entity_detection_statistics{$type}{correct});
	printf "  %3.3s%8d%6d%9d%6d%6d%11.1f%6.1f%9.1f%6.1f%6.1f\n", $type,
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{fa},
	$entity_detection_statistics{$type}{miss},
	$entity_detection_statistics{$type}{error},
	$entity_detection_statistics{$type}{correct},
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{fa}/max($total,1),
	100*$entity_detection_statistics{$type}{miss}/max($total,1),
	100*$entity_detection_statistics{$type}{error}/max($total,1),
	100*$entity_detection_statistics{$type}{correct}/max($total,1);
    }
}

#################################

sub score_EDT_attribute_recognition {

    my ($type, $ref_type);

    foreach $ref_type (keys %entity_type_statistics) {
	foreach $type (keys %entity_type_statistics) {
	    next if $type eq "total";
	    $entity_type_statistics{$ref_type}{total} += $entity_type_statistics{$ref_type}{$type};
	}
    }

    print "\nEntity Type confusion matrix (for mapped entities):\n",
    "               ___________count___________        __________percent__________\n",
    "  ref \\ tst:  ";
    foreach $type (@entity_types) {
	last if $type eq "total";
	printf " %3.3s  ", $type;
    }
    print "   ";
    foreach $type (@entity_types) {
	last if $type eq "total";
	printf "   %3.3s", $type;
    }
    print "\n";
    foreach $ref_type (@entity_types) {
	last if $ref_type eq "total";
	printf "  %3.3s       ", $ref_type;
	foreach $type (@entity_types) {
	    printf "%6d", $entity_type_statistics{$ref_type}{$type}
	    unless $type eq "total";
	}
	print "     ";
	foreach $type (@entity_types) {
	    printf "%6.1f", 100*$entity_type_statistics{$ref_type}{$type} /
		max($entity_type_statistics{$ref_type}{total},1)
		    unless $type eq "total";
	}
	print "\n";
    }

    foreach $type (keys %name_detection_statistics) {
	next if $type eq "total";
	$name_detection_statistics{total}{miss} += $name_detection_statistics{$type}{miss};
	$name_detection_statistics{total}{fa} += $name_detection_statistics{$type}{fa};
	$name_detection_statistics{total}{correct} += $name_detection_statistics{$type}{correct};
	$name_detection_statistics{total}{error} += $name_detection_statistics{$type}{error};
    }

    print "\nName Detection and Extent Recognition statistics (for mapped entities):\n",
    " ref     ____________count______________       ____________percent____________\n",
    " entity  Detection     Extent_Recognition      Detection     Extent_Recognition\n",
    " type    miss    fa     miss   err  corr       miss    fa     miss   err  corr\n";
    foreach $type (@entity_types) {
	my $total = ($name_detection_statistics{$type}{miss} +
		     $name_detection_statistics{$type}{error} +
		     $name_detection_statistics{$type}{correct});
	printf "  %3.3s%8d%6d%9d%6d%6d%11.1f%6.1f%9.1f%6.1f%6.1f\n", $type,
	$name_detection_statistics{$type}{miss},
	$name_detection_statistics{$type}{fa},
	$name_detection_statistics{$type}{miss},
	$name_detection_statistics{$type}{error},
	$name_detection_statistics{$type}{correct},
	100*$name_detection_statistics{$type}{miss}/max($total,1),
	100*$name_detection_statistics{$type}{fa}/max($total,1),
	100*$name_detection_statistics{$type}{miss}/max($total,1),
	100*$name_detection_statistics{$type}{error}/max($total,1),
	100*$name_detection_statistics{$type}{correct}/max($total,1);
    }
}

#################################

sub score_EDT_mention_recognition {

    my $type;

    foreach $type (keys %mention_detection_statistics) {
	next if $type eq "total";
	$mention_detection_statistics{total}{miss} += $mention_detection_statistics{$type}{miss};
	$mention_detection_statistics{total}{fa} += $mention_detection_statistics{$type}{fa};
	$mention_detection_statistics{total}{correct} += $mention_detection_statistics{$type}{correct};
	$mention_detection_statistics{total}{error} += $mention_detection_statistics{$type}{error};
    }

    print "\nMention Detection and Extent Recognition statistics (for mapped entities):\n",
    " ref     ____________count______________       ____________percent____________\n",
    " mention Detection     Extent_Recognition      Detection     Extent_Recognition\n",
    " type    miss    fa     miss   err  corr       miss    fa     miss   err  corr\n";
    foreach $type (@mention_types) {
	my $total = ($mention_detection_statistics{$type}{miss} +
		     $mention_detection_statistics{$type}{error} +
		     $mention_detection_statistics{$type}{correct});
	printf "  %3.3s%8d%6d%9d%6d%6d%11.1f%6.1f%9.1f%6.1f%6.1f\n", $type,
	$mention_detection_statistics{$type}{miss},
	$mention_detection_statistics{$type}{fa},
	$mention_detection_statistics{$type}{miss},
	$mention_detection_statistics{$type}{error},
	$mention_detection_statistics{$type}{correct},
	100*$mention_detection_statistics{$type}{miss}/max($total,1),
	100*$mention_detection_statistics{$type}{fa}/max($total,1),
	100*$mention_detection_statistics{$type}{miss}/max($total,1),
	100*$mention_detection_statistics{$type}{error}/max($total,1),
	100*$mention_detection_statistics{$type}{correct}/max($total,1);
    }
}

#################################

sub entity_mention_type {

    my ($entity) = @_;
    my ($mentions, $mention);

#determine the type of mentions for this entity
    my $mention_type = "PRONOUN";
    $mentions = $entity->{mentions};
    foreach $mention (@$mentions) {
	$mention_type = "NAME" if ($mention->{TYPE} eq "NAME");
	$mention_type = $mention->{TYPE} if $mention_type eq "PRONOUN";
    }
    return $mention_type;
}

#################################

sub entity_count_type {

    my ($entity) = @_;
    my ($count, $count_type);

#determine the number of mentions for this entity
    $count = @{$entity->{mentions}};
    return $count le 1 ? $count_types[0] :
	   $count le 2 ? $count_types[1] :
	   $count le 4 ? $count_types[2] :
	   $count le 8 ? $count_types[3] : $count_types[4];
}

#################################

sub tabulate_EDT_scoring_statistics {

    my ($ref_entities, $tst_entities) = @_;
    my ($ref_entity, $tst_entity);
    my ($ref_mentions, $tst_mentions);
    my ($ref_mention, $tst_mention);
    my ($ref_names, $tst_names);
    my ($ref_name, $tst_name);
    my ($overlap, $max_overlap);
    my $max_extent_diff;

#select criterion for matching extents
    if ($document->{TYPE} eq "text") {
	$max_extent_diff = $max_diff_chars;
    }
    elsif ($document->{TYPE} eq "audio") {
	$max_extent_diff = $max_diff_time;
    }
    elsif ($document->{TYPE} eq "image") {
	$max_extent_diff = $max_diff_xy;
    }
    else {
	die "\n\nFATAL ERROR in tabulate_EDT_scoring_statistic:  ".
	    "source type '$document->{TYPE}' is unknown\n\n";
    }

#accumulate recognition statistics over all entities
    foreach $ref_entity (@$ref_entities) {

	my $mention_type = entity_mention_type ($ref_entity);
	my $count_type = entity_count_type ($ref_entity);

#measure performance only for those entities that are correctly detected
	if ($ref_entity->{MAP}) {

#update entity detection statistics
	    if ($ref_entity->{TYPE} eq $ref_entity->{MAP}{TYPE}) {
		$entity_detection_statistics{$ref_entity->{TYPE}}{correct}++;
		$entity_detection_statistics{$mention_type}{correct}++;
		$entity_detection_statistics{$count_type}{correct}++;
	    }
	    else {
		$entity_detection_statistics{$ref_entity->{TYPE}}{error}++;
		$entity_detection_statistics{$mention_type}{error}++;
		$entity_detection_statistics{$count_type}{error}++;
	    }

#update entity type statistics
	    $entity_type_statistics{$ref_entity->{TYPE}}{$ref_entity->{MAP}{TYPE}}++;

#update entity name statistics
	    $ref_names = $ref_entity->{names};
	    $tst_names = $ref_entity->{MAP}{names};
	    foreach $ref_name (@$ref_names) {
		$max_overlap = $min_overlap;
		foreach $tst_name (@$tst_names) {
		    $overlap = span_overlap($ref_name->{locator}, $tst_name->{locator});
		    if ($overlap > $max_overlap) {
			$max_overlap = $overlap;
			$ref_name->{MAP} = $tst_name;
		    }
		}
		if ($ref_name->{MAP}) {
		    $tst_name = $ref_name->{MAP};
		    $tst_name->{MAP} = $ref_name;
		    extent_mismatch($ref_name->{locator},
					   $tst_name->{locator})
			<= $max_extent_diff ?
			    $name_detection_statistics{$ref_entity->{TYPE}}{correct}++ :
				$name_detection_statistics{$ref_entity->{TYPE}}{error}++;
		}
		else {
		    $name_detection_statistics{$ref_entity->{TYPE}}{miss}++;
		}
	    }
	    foreach $tst_name (@$tst_names) {
		if (not $tst_name->{MAP}) {
		    $name_detection_statistics{$ref_entity->{TYPE}}{fa}++;
		}
	    }

#update mention statistics
	    $ref_mentions = $ref_entity->{mentions};
	    $tst_mentions = $ref_entity->{MAP}->{mentions};
	    foreach $ref_mention (@$ref_mentions) {
		foreach $tst_mention (@$tst_mentions) {
		    if (mention_mapping_score($ref_mention, $tst_mention)) {
			$ref_mention->{MAP} = $tst_mention;
			$tst_mention->{MAP} = $ref_mention;
		    }
		}
		if ($ref_mention->{MAP}) {
		    extent_mismatch ($ref_mention->{extent}{locator},
					    $ref_mention->{MAP}->{extent}{locator})
			    <= $max_extent_diff ?
				$mention_detection_statistics{$ref_mention->{TYPE}}{correct}++ :
				    $mention_detection_statistics{$ref_mention->{TYPE}}{error}++;
		}
		else {
		    $mention_detection_statistics{$ref_mention->{TYPE}}{miss}++;
		}
	    }
	    foreach $tst_mention (@$tst_mentions) {
		if (not $tst_mention->{MAP}) {
		    $mention_detection_statistics{$tst_mention->{TYPE}}{fa}++;
		}
	    }
	}
	else {
	    $entity_detection_statistics{$ref_entity->{TYPE}}{miss}++;
	    $entity_detection_statistics{$mention_type}{miss}++;
	    $entity_detection_statistics{$count_type}{miss}++;
	}
    }
    foreach $tst_entity (@$tst_entities) {
	if (not $tst_entity->{MAP}) {
	    $entity_detection_statistics{$tst_entity->{TYPE}}{fa}++;
	    $entity_detection_statistics{entity_mention_type($tst_entity)}{fa}++;
	    $entity_detection_statistics{entity_count_type($tst_entity)}{fa}++;
	}
    }
}

#################################

sub initialize_statistics {

    my ($ref_type, $tst_type, $err_type, $men_type);

    foreach $ref_type (@entity_types) {
	foreach $err_type (@error_types) {
	    $entity_detection_statistics{$ref_type}{$err_type} = 0;
	    $name_detection_statistics{$ref_type}{$err_type} = 0;
	}
	foreach $tst_type (@entity_types) {
	    $entity_type_statistics{$ref_type}{$tst_type} = 0;
	}
    }
    foreach $ref_type (@count_types) {
	foreach $err_type (@error_types) {
	    $entity_detection_statistics{$ref_type}{$err_type} = 0;
	}
    }
    foreach $men_type (@mention_types) {
	foreach $err_type (@error_types) {
	    $mention_detection_statistics{$men_type}{$err_type} = 0;
	    $entity_detection_statistics{$men_type}{$err_type} = 0;
	}
    }
}

#################################

sub print_scoring_details {

    my ($ref_entities, $tst_entities) = @_;
    my ($ref_entity, $tst_entity);
    my ($ref_mentions, $tst_mentions);
    my ($ref_mention, $tst_mention);
    my ($ref_names, $tst_names);
    my ($ref_name, $tst_name);
    my $output;
    my $separator = "-" x 18 . "\n";
    my $print_data;
    my $max_extent_diff;

#select criterion for matching extents
    if ($document->{TYPE} eq "text") {
	$max_extent_diff = $max_diff_chars;
    }
    elsif ($document->{TYPE} eq "audio") {
	$max_extent_diff = $max_diff_time;
    }
    elsif ($document->{TYPE} eq "image") {
	$max_extent_diff = $max_diff_xy;
    }
    else {
	die "\n\nFATAL ERROR in tabulate_EDT_scoring_statistic:  ".
	    "source type '$document->{TYPE}' is unknown\n\n";
    }

    foreach $ref_entity (@$ref_entities) {
	$output = $separator;
	$print_data = $print_all_data;
	if ($ref_entity->{MAP}) {
	    if ($ref_entity->{TYPE} eq $ref_entity->{MAP}->{TYPE}) {
		$output .= "    ref entity $ref_entity->{ID} ($ref_entity->{TYPE})\n";
		$output .= "    tst entity $ref_entity->{MAP}->{ID} ($ref_entity->{MAP}->{TYPE})\n";
	    }
	    else {
		$print_data ||= $print_err_data || $print_ent_data;
		$output .= ">>> ref entity $ref_entity->{ID} ($ref_entity->{TYPE}) -- ENTITY TYPE MISMATCH\n";
		$output .= ">>> tst entity $ref_entity->{MAP}->{ID} ($ref_entity->{MAP}->{TYPE}) -- ENTITY TYPE MISMATCH\n";
	    }

	    $ref_mentions = $ref_entity->{mentions};
	    $tst_mentions = $ref_entity->{MAP}->{mentions};
	    foreach $ref_mention (@$ref_mentions) {
		if ($ref_mention->{MAP}) {
		    if ($ref_mention->{TYPE} eq $ref_mention->{MAP}->{TYPE}) {
			$output .= "      ref mention = '$ref_mention->{head}{text}'";
			$output .= " ($ref_mention->{TYPE})\n";
		    }
		    else {
			$print_data ||= $print_err_data || $print_men_data;
			$output .= ">>>   ref mention = '$ref_mention->{head}{text}'";
			$output .= " ($ref_mention->{TYPE}) -- MENTION TYPE MISMATCH\n";
			$output .= ">>>   tst mention = '$ref_mention->{MAP}->{head}{text}'";
			$output .= " ($ref_mention->{MAP}->{TYPE}) -- MENTION TYPE MISMATCH\n";
		    }
		    if (extent_mismatch ($ref_mention->{extent}{locator},
						$ref_mention->{MAP}->{extent}{locator})
			> $max_extent_diff) {
			$print_data ||= $print_err_data || $print_men_data;
 			$output .= ">>>    ref mention extent = '$ref_mention->{extent}{text}'";
			$output .= " -- MENTION EXTENT MISMATCH\n";
			$output .= ">>>    tst mention extent = '$ref_mention->{MAP}->{extent}{text}'";
			$output .= " -- MENTION EXTENT MISMATCH\n";
		    }
		}
 		else {
		    $print_data ||= $print_err_data || $print_men_data;
		    $output .= ">>>   ref mention = '$ref_mention->{head}{text}'";
		    $output .= " ($ref_mention->{TYPE}) -- NO MATCHING TST MENTION\n";
		}
	    }
	    foreach $tst_mention (@$tst_mentions) {
		if (not $tst_mention->{MAP}) {
		    $print_data ||= $print_err_data || $print_men_data;
		    $output .= ">>>   tst mention = '$tst_mention->{head}{text}'";
		    $output .= " ($tst_mention->{TYPE}) -- NO MATCHING REF MENTION\n";
		}
	    }
	    $ref_names = $ref_entity->{names};
	    foreach $ref_name (@$ref_names) {
		if ($ref_name->{MAP}) {
		    if (extent_mismatch($ref_name->{locator},
					       $ref_name->{MAP}{locator})
			<= $max_extent_diff) {
			$output .= "         ref name = '$ref_name->{text}'\n";
		    }
		    else {
			$print_data ||= $print_err_data || $print_nam_data;
			$output .= ">>>       ref name extent = '$ref_name->{text}'";
			$output .= " -- NAME EXTENT MISMATCH\n";
			$output .= ">>>       tst name extent = '$ref_name->{MAP}{text}'";
			$output .= " -- NAME EXTENT MISMATCH\n";
		    }
		}
		else {
		    $print_data ||= $print_err_data || $print_nam_data;
		    $output .= ">>>      ref name = '$ref_name->{text}'";
		    $output .= " -- NO MATCHING TST NAME\n";
		}
	    }
	    $tst_names = $ref_entity->{MAP}{names};
	    foreach $tst_name (@$tst_names) {
		if (not $tst_name->{MAP}) {
		    $print_data ||= $print_err_data || $print_nam_data;
		    $output .= ">>>      tst name = '$tst_name->{text}'";
		    $output .= " -- NO MATCHING REF NAME\n";
		}
	    }
	}
	else {
	    $print_data ||= $print_err_data || $print_ent_data;
	    $output .= ">>> ref entity $ref_entity->{ID} ($ref_entity->{TYPE}) -- NO MATCHING TST ENTITY\n";
	    $ref_mentions = $ref_entity->{mentions};
	    foreach $ref_mention (@$ref_mentions) {
		$output .= "      ref mention = '$ref_mention->{head}{text}'";
		$output .= " ($ref_mention->{TYPE}) -- unscored\n";
	    }
	}
	print $output if $print_data;
    }
    foreach $tst_entity (@$tst_entities) {
	$print_data = $print_all_data;
	if (not $tst_entity->{MAP}) {
	    $output = $separator;
	    $print_data ||= $print_all_data || $print_err_data || $print_ent_data;
	    $output .= ">>> tst entity $tst_entity->{ID} ($tst_entity->{TYPE}) -- NO MATCHING REF ENTITY\n";
	    $tst_mentions = $tst_entity->{mentions};
	    foreach $tst_mention (@$tst_mentions) {
		$output .= "      tst mention = '$tst_mention->{head}{text}'";
		$output .= " ($tst_mention->{TYPE}) -- unscored\n";
	    }
	    print $output if $print_data;
	}
    }
}

#################################

sub map_EDT_entities {

    my ($ref_entities, $tst_entities) = @_;
    my ($ref_entity, $tst_entity);
    my (@ref_cohorts, @tst_cohorts);
    my $null_candidate = {entity => {}, score => 0};

#map all reference entities
    foreach $ref_entity (@$ref_entities) {

#unless the entity has already been mapped
	unless (exists $ref_entity->{MAP}) {
	    print "-" x 40, "\n" if $print_mapping_detail;
    
#determine the cohort set of possible mappings for this ref entity
	    tag_cohorts ($ref_entity, $ref_entities, $tst_entities);
	    @ref_cohorts = collect_cohorts ($ref_entities);
	    @tst_cohorts = collect_cohorts ($tst_entities);

#order tst entities so that those with the best potential match come first
	    foreach $ref_entity (@ref_cohorts) {
		push my (@map_candidates), $null_candidate;
		foreach $tst_entity (@tst_cohorts) {
		    my $score = entity_mapping_score ($ref_entity, $tst_entity);
		    push @map_candidates, {entity => $tst_entity, score => $score}
		        if ($score > 0);
		}
		my @tst_ordered = sort {$b->{score} <=> $a->{score}} @map_candidates;
		$ref_entity->{tst_ordered} = [@tst_ordered];
		$ref_entity->{max_score} = $ref_entity->{tst_ordered}[0]->{score};
	    }
	    
#order ref entities so that those with the largest potential score come first
	    @ref_cohorts = sort {$b->{max_score} <=> $a->{max_score}} @ref_cohorts;
	    my $score_bound = 0;
	    foreach $ref_entity (reverse @ref_cohorts) {
		$ref_entity->{score_bound} = $score_bound;
		$score_bound += $ref_entity->{max_score};
	    }

#perform exhaustive search for optimum mapping over the cohort set
	    print_mapping_header (\@ref_cohorts, \@tst_cohorts)
		if $print_mapping_detail;
	    find_optimum_mapping ($best_score=0, \@ref_cohorts,
				  \@tst_cohorts, \@ref_cohorts);

#create test entity backpointers for all ref-tst mappings
	    foreach $ref_entity (@ref_cohorts) {
		if ($ref_entity->{MAP}) {
		    $ref_entity->{MAP}->{MAP} = $ref_entity;
		}
		else {
		    $ref_entity->{MAP} = 0;
		}
	    }
	}
    }
}

#################################

sub find_optimum_mapping {

    my ($prior_score, $ref_entities, $tst_entities, $ref_cohorts) = @_;
    my ($ref_entity, $tst_entity);
    my (@ref_remaining) = @$ref_entities;

#map the next ref entity
    if ($ref_entity = shift (@ref_remaining)) {
	return if ($prior_score+$ref_entity->{max_score} <= $best_score);
	
#now try all mappings
	my @tst_ordered = @{$ref_entity->{tst_ordered}};
	while (my $map_candidate = shift @tst_ordered) {

#each test entity may map to only one ref entity
	    $tst_entity = $map_candidate->{entity};
	    unless ($tst_entity->{MAP}) {
		
#try this mapping
		$tst_entity->{MAP} = $ref_entity if $tst_entity->{ID};
		find_optimum_mapping ($map_candidate->{score}+$prior_score,
				      \@ref_remaining, $tst_entities, $ref_cohorts);
		delete $tst_entity->{MAP};
	    }
	}
    }
	
#all reference entities have been mapped
    else {
	print_mapping_detail ($prior_score, $tst_entities)
	    if $print_mapping_detail;
	if ($prior_score > $best_score) { #keep better mapping
	    $best_score = $prior_score;
	    foreach my $entity (@$ref_cohorts) { #delete old mapping
		delete $entity->{MAP};
	    }
	    foreach my $entity (@$tst_entities) { #create new mapping
		$entity->{MAP}->{MAP} = $entity
		    if ($entity->{MAP});
	    }
	}
    }
}

#################################

sub print_mapping_header {

    my ($ref_cohorts, $tst_cohorts) = @_;

    foreach my $entity (@$ref_cohorts) {
	print "ref=$entity->{ID}, max_score=$entity->{max_score}, ".
	    "score_bound=$entity->{score_bound}\n";
	foreach my $map_candidate (@{$entity->{tst_ordered}}) {
	    print "   tst=$map_candidate->{entity}->{ID}, score=$map_candidate->{score}\n"
		if ($map_candidate->{score} > 0);
	}
    }
    print "REF cohort:";
    foreach my $entity (@$ref_cohorts) {
	print " ", substr($entity->{ID}, length($entity->{ID})-2, 2);
    }
    print "\nTST cohort:";
    foreach my $entity (@$tst_cohorts) {
	print " ", substr($entity->{ID}, length($entity->{ID})-2, 2);
    }
    print " score\n";
}

#################################

sub print_mapping_detail {

    my ($score, $tst_entities) = @_;

    print "       ref:";
    foreach my $entity (@$tst_entities) {
	$entity->{MAP} ?
	    print " ", substr($entity->{MAP}->{ID}, length($entity->{MAP}->{ID})-2, 2) :
		print "   ";
    }
    print "  $score",	($score > $best_score) ? "<=max\n" : "\n";
}

#################################

sub entity_mapping_score {

#N.B.  The entity mapping score must be 0 if tst doesn't match ref.

    my ($ref_entity, $tst_entity) = @_;
    my ($ref_mentions, $tst_mentions);
    my ($ref_mention, $tst_mention);

#check for null entity
    return 0 if not ($ref_entity and $tst_entity);

    my $score = $entity_mapping_score{$ref_entity}{$tst_entity};

#use table-lookup for precomputed scores
    return $score if defined $score;

#find the matching test mention for each reference mention
    $score = 0;
    $ref_mentions = $ref_entity->{mentions};
    $tst_mentions = $tst_entity->{mentions};
    foreach $ref_mention (@$ref_mentions) {
	foreach $tst_mention (@$tst_mentions) {
	    $score += mention_mapping_score 
		($ref_mention, $tst_mention);
	}
    }
    $entity_mapping_score{$ref_entity}{$tst_entity} = $score;
    return $score;
}
      
#################################

sub mention_mapping_score {

#N.B.  The mention mapping score must be 0 if tst doesn't match ref.

    my ($ref_mention, $tst_mention) = @_;
    my $mention_score;

#a test mention matches a ref mention if and only if:
#  the test mention points to the head of the ref mention
#--------
#The algorithm for determining whether the test mention "points to"
#the reference mention is to compute the mid-point of the test mention
#and to require this mid-point to lie within the span of the reference
#mention.

    $mention_score = 0;
    if (span_overlap($ref_mention->{head}{locator}, $tst_mention->{head}{locator})
	>= $min_overlap) {
	$mention_score = 
	    $ref_mention->{TYPE} eq "NAME" ? $nam_match :
		$ref_mention->{TYPE} eq "NOMINAL" ? $nom_match :
		    $ref_mention->{TYPE} eq "PRONOUN" ? $pro_match : 0;
    }
    return $mention_score;
}
      
#################################
      
sub collect_cohorts {

    my ($entities) = @_;
    my ($entity, @cohorts);

    foreach $entity (@$entities) {
	if (exists($entity->{cohort})) {
	    push @cohorts, $entity;
	    delete $entity->{cohort};
	}
    }
    return @cohorts;
}

#################################

sub tag_cohorts {

    my ($ent, $ent_set, $map_set) = @_;
    my $ent_map;

    $ent->{cohort} = "";
    foreach $ent_map (@$map_set) {
	unless (exists($ent_map->{cohort})) {
	    if (entity_mapping_score ($ent, $ent_map)) {
		tag_cohorts ($ent_map, $map_set, $ent_set);
	    }
	}
    }
}

#################################

sub span_overlap {

#This subroutine returns the minimum mutual overlap between two spans.

    my ($ref_locator, $tst_locator) = @_;
    if ($document->{TYPE} eq "text") {
	return text_span_overlap ($ref_locator, $tst_locator);
    }
    elsif ($document->{TYPE} eq "audio") {
	return audio_span_overlap ($ref_locator, $tst_locator);
    }
    elsif ($document->{TYPE} eq "image") {
	return image_span_overlap ($ref_locator, $tst_locator);
    }
    die "\n\nFATAL ERROR in span_overlap:  source type '$document->{TYPE}' is unknown\n\n";
}

#################################

sub text_span_overlap {

#This subroutine returns the minimum mutual overlap between two spans.

    my ($ref_locator, $tst_locator) = @_;
    my ($ref_start, $ref_end, $tst_start, $tst_end);
    my ($ref_overlap, $tst_overlap);
    
    $ref_start = $ref_locator->{start};
    $ref_end = $ref_locator->{end};
    $tst_start = $tst_locator->{start};
    $tst_end = $tst_locator->{end};

#compute overlap of ref_locator by tst_locator
    $tst_overlap = $tst_start > $ref_start ?
	$ref_end - $tst_start + 1 :
	    $tst_end - $ref_start + 1;
    $tst_overlap /= ($ref_end - $ref_start + 1);

#compute overlap of tst_locator by ref_locator
    $ref_overlap = $ref_start > $tst_start ?
	$tst_end - $ref_start + 1 :
	    $ref_end - $tst_start + 1;
    $ref_overlap /= ($tst_end - $tst_start + 1);

#return the minimum mutual overlap
    return min($tst_overlap, $ref_overlap);
}

#################################

sub audio_span_overlap {

#This subroutine returns the minimum mutual overlap between two spans.

    my ($ref_locator, $tst_locator) = @_;
    my ($ref_start, $ref_end, $tst_start, $tst_end);
    my ($ref_overlap, $tst_overlap);
    my $nominal_frame_period = 0.01;
    
    $ref_start = $ref_locator->{tstart};
    $ref_end = $ref_start + $ref_locator->{tdur};
    $tst_start = $tst_locator->{tstart};
    $tst_end = $tst_start + $tst_locator->{tdur};

#compute overlap of ref_locator by tst_locator
    $tst_overlap = ($tst_start > $ref_start ?
		    $ref_end - $tst_start : $tst_end - $ref_start)
	+ $nominal_frame_period;
    $tst_overlap /= ($ref_end - $ref_start + $nominal_frame_period);

#compute overlap of tst_locator by ref_locator
    $ref_overlap = ($ref_start > $tst_start ?
		    $tst_end - $ref_start : $ref_end - $tst_start)
	+ $nominal_frame_period;
    $ref_overlap /= ($tst_end - $tst_start + $nominal_frame_period);

#return the minimum mutual overlap
    return min($tst_overlap, $ref_overlap);
}

#################################

sub image_span_overlap {

#This subroutine returns the minimum mutual overlap between two image spans.

    my ($ref_boxes, $tst_boxes) = @_;
    my ($ref_box, $tst_box);
    my $nominal_pixel_size = 1; # = 1 cm???!!! no -- current scale is pixels rather than cm

    my $ref_area = 0;
    foreach $ref_box (@$ref_boxes) {
	$ref_area += ($ref_box->{width} + $nominal_pixel_size)
	    * ($ref_box->{height} + $nominal_pixel_size);
    }

    my $tst_area = 0;
    foreach $tst_box (@$tst_boxes) {
	$tst_area += ($tst_box->{width} + $nominal_pixel_size)
	    * ($tst_box->{height} + $nominal_pixel_size);
    }
    
    my $mutual_area;
    my $cum_mutual_area = 0;
    foreach $ref_box (@$ref_boxes) {
	my $max_mutual_area = 0;
	foreach $tst_box (@$tst_boxes) {
	    next if ($ref_box->{page} ne $tst_box->{page});
	    my $x1 = max($ref_box->{x_start},
			 $tst_box->{x_start});
	    my $x2 = min($ref_box->{x_start}+$ref_box->{width},
			 $tst_box->{x_start}+$tst_box->{width});
	    next if ($x1 > $x2);
	    my $y1 = max($ref_box->{y_start},
			 $tst_box->{y_start});
	    my $y2 = min($ref_box->{y_start}+$ref_box->{height},
			 $tst_box->{y_start}+$tst_box->{height});
	    next if ($y1 > $y2);
	    $mutual_area = ($x2 - $x1 + $nominal_pixel_size)
		* ($y2 - $y1 + $nominal_pixel_size);
	    $max_mutual_area = $mutual_area
		if $mutual_area > $max_mutual_area;
	}
	$cum_mutual_area += $max_mutual_area;
    }

#return the minimum mutual overlap
    return $cum_mutual_area /
	max($ref_area,$tst_area);
}

#################################

sub extent_mismatch {

#This subroutine returns the maximum mismatch in the
#extent of two locators

    my ($ref_locator, $tst_locator) = @_;
    if ($document->{TYPE} eq "text") {
	return text_extent_mismatch ($ref_locator, $tst_locator);
    }
    elsif ($document->{TYPE} eq "audio") {
	return audio_extent_mismatch ($ref_locator, $tst_locator);
    }
    elsif ($document->{TYPE} eq "image") {
	return image_extent_mismatch ($ref_locator, $tst_locator);
    }
    die "\n\nFATAL ERROR in extent_mismatch:  source type '$document->{TYPE}' is unknown\n\n";
}

#################################

sub text_extent_mismatch {

#This subroutine returns the maximum mismatch in the
#character extent of two text streams

    my ($ref_locator, $tst_locator) = @_;
    my $extent_mismatch = 0;

    $extent_mismatch =
	max(abs($ref_locator->{start} - $tst_locator->{start}),
	    abs($ref_locator->{end} - $tst_locator->{end}));
}

#################################

sub audio_extent_mismatch {

#This subroutine returns the maximum mismatch in the
#time extent of two audio signals

    my ($ref_locator, $tst_locator) = @_;
    my $extent_mismatch = 0;
    my ($ref_start, $ref_end, $tst_start, $tst_end);

    $ref_start = $ref_locator->{tstart};
    $tst_start = $tst_locator->{tstart};
    $ref_end = $ref_start + $ref_locator->{tdur};
    $tst_end = $tst_start + $tst_locator->{tdur};

    $extent_mismatch =
	max(abs($ref_start - $tst_start),
	    abs($ref_end - $tst_end));
}

#################################

sub image_extent_mismatch {

#This subroutine returns the maximum mismatch in the
#spatial extent of two images.

    my ($ref_boxes, $tst_boxes) = @_;
    my ($ref_box, $tst_box);
    my ($x_mismatch, $y_mismatch);
    my $huge_mismatch = 9999999;
    my $extent_mismatch = 0;

    foreach $ref_box (@$ref_boxes) {
	$x_mismatch = $y_mismatch = $huge_mismatch;
	foreach $tst_box (@$tst_boxes) {
	    next if ($ref_box->{page} ne $tst_box->{page});
	    $x_mismatch = min($x_mismatch,
			      abs($ref_box->{x_start} - $tst_box->{x_start}),
			      abs($ref_box->{x_start}+$ref_box->{width} -
				  $tst_box->{x_start}+$tst_box->{width}));
	    $y_mismatch = min($y_mismatch,
			      abs($ref_box->{y_start} - $tst_box->{y_start}),
			      abs($ref_box->{y_start}+$ref_box->{height} -
				  $tst_box->{y_start}+$tst_box->{height}));
	}
	$extent_mismatch = max($extent_mismatch, $x_mismatch, $y_mismatch);
    }
    return $extent_mismatch;
}

#################################

sub max {

    my ($max, $x);

    $max = shift @_;
    foreach $x (@_) {
	if ($x > $max) {$max = $x}
    }
    return $max;
}

#################################

sub min {

    my ($min, $x);

    $min = shift @_;
    foreach $x (@_) {
	if ($x < $min) {$min = $x}
    }
    return $min;
}

#################################

sub get_EDT_data {

    my ($docs, $file) = @_;
    my ($line, $name);
    my ($data, $tag, $span, $type);
    my ($source, $doc_id, %src_doc, @entities);
    my $ndocs = 0;

#read data from file
    open (FILE, $file) or die "\nUnable to open EDT output file '$input_file'", $usage;
    while ($line=<FILE>){
	$data .= $line;
    }
    close (FILE);

#get source file information
    ($tag, $span) = 
	extract_sgml_tag_and_span ("source_file", $data);
    die "\n\nFATAL INPUT ERROR:  no source file tag attribute '$name' in '$input_file'\n\n"
	unless (($type) = extract_sgml_tag_attribute ($name="TYPE", $tag));
    die "\n\nFATAL INPUT ERROR:  no source file tag attribute '$name' in '$input_file'\n\n"
	unless (($source) = extract_sgml_tag_attribute ($name="URI", $tag));

#get document data for all documents in the source file
    while (($tag, $span, $data) = 
	   extract_sgml_tag_and_span ("document", $data))
    {
	die "\n\nFATAL INPUT ERROR:  no document ID found in tag '$tag' in file '$input_file'\n\n"
	    unless (($doc_id) = extract_sgml_tag_attribute ("DOCID", $tag));
	die "\n\nFATAL INPUT ERROR:  document ID '$doc_id' in file '$input_file' is not unique\n\n"
	    if defined ($docs->{$doc_id});
	$input_doc = $doc_id;
	$src_doc{TYPE} = $type;
	$src_doc{FILE} = $source;
	$document->{TYPE} = $src_doc{TYPE};
	$src_doc{entities} = [get_entities ($span)];
	($docs->{$doc_id}) = {%src_doc};
	$ndocs++;
    }
    print "\n\nWARNING:  file '$input_file' contains no documents\n\n"
	if ($ndocs eq 0);
}

#################################

sub get_entities {

#extract information for all entities in the document

    my ($span) = @_;
    my (%entity, @entities);
    my ($tag, $data, $type);
    my $nentities = 0;

    while (($tag, $data, $span) =
	   extract_sgml_tag_and_span ("entity", $span))
    {
	die "\n\nFATAL INPUT ERROR:  no entity ID found for an entity in file '$input_file':\n\n".
	    "Entity tag is '$tag'\nEntity data is '$data'\n"
		unless (($input_entity) = extract_sgml_tag_attribute ("ID", $tag));
	die "\n\nFATAL INPUT ERROR for entity '$input_entity' in document '$input_doc'".
	    " in file '$input_file':\n  no entity TYPE found.  Entity data is '$data'\n\n"
		unless (($type) = extract_sgml_span ("entity_type", $data));

	$entity{ID} = $input_entity;
	$entity{TYPE} = $type eq "GSP" ? "GPE" : $type;
	$entity{mentions} = [get_entity_mentions ($data)];
	$entity{names} = [get_entity_names ($data)];

	push @entities, {%entity};
	$nentities++;
    }
    undef(%entity_mapping_score);
    print "\n\nWARNING:  document '$input_doc' in file '$input_file' contains no entities\n\n"
	if ($nentities eq 0);
    
    return @entities;
}

#################################

sub get_entity_mentions {

    my ($mention_data) = @_;
    my (%mention, @mentions);
    my ($tag, $data, $type);
    my ($head, $extent);
    my (%head, %extent);
    my $text;
    my $nmentions = 0;
    
    while (($tag, $data, $mention_data) =
	   extract_sgml_tag_and_span ("entity_mention", $mention_data))
    {
	die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	    "No mention TYPE found in tag '$tag'\n\n"
		unless (($type) = extract_sgml_tag_attribute ("TYPE", $tag));
	die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	    "No mention head found in data ($data)\n\n"
		unless (($head) = extract_sgml_span ("head", $data));
	die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	    "No mention extent found in data ($data)\n\n"
		unless (($extent) = extract_sgml_span ("extent", $data));

	$mention{"TYPE"} = $type;
	$head{text} = extract_sgml_tag_attribute ("string", $head);
	$extent{text} = extract_sgml_tag_attribute ("string", $extent);
	if ($document->{TYPE} eq "text") {
	    ($head{locator}) = get_text_locator ($head);
	    ($extent{locator}) = get_text_locator ($extent);
	}
	elsif ($document->{TYPE} eq "audio") {
	    ($head{locator}) = get_audio_locator ($head);
	    ($extent{locator}) = get_audio_locator ($extent);
	}
	elsif ($document->{TYPE} eq "image") {
	    ($head{locator}) = [get_image_locator ($head)];
	    ($extent{locator}) = [get_image_locator ($extent)];
	}
	else {
	    die "\n\nFATAL INPUT ERROR for document '$input_doc' in file '$input_file':\n".
		"No locator read routine for this document type ($document->{TYPE})\n\n";
	}
	$mention{head} = {%head};
	$mention{extent} = {%extent};
	push @mentions, {%mention};
	$nmentions++;
    }
    die "\n\nFATAL INPUT ERROR for entity '$input_entity' in document '$input_doc'".
	"in file '$input_file':\n  entity contains no mentions\n\n"
	    if ($nmentions eq 0);
    
    return @mentions;
}

#################################
    
sub get_entity_names {

    my ($name_data) = @_;
    my $span;
    my (%name, @names);
    my ($tag, $data, $start, $end, $text);
    
    ($span) = extract_sgml_span ("entity_attributes", $name_data);

    while ($span and (($tag, $data, $span) = 
	   extract_sgml_tag_and_span ("name", $span)))
    {
	$name{text} = extract_sgml_tag_attribute ("string", $data);

	if ($document->{TYPE} eq "text") {
	    ($name{locator}) = get_text_locator ($data);
	}
	elsif ($document->{TYPE} eq "audio") {
	    ($name{locator}) = get_audio_locator ($data);
	}
	elsif ($document->{TYPE} eq "image") {
	    ($name{locator}) = [get_image_locator ($data)];
	}
	else {
	    die "\n\nFATAL INPUT ERROR for document '$input_doc' in file '$input_file':\n".
		"No locator read routine for this document type ($document->{TYPE})\n\n";
	}

	push @names, {%name};
    }
    return @names;
}

#################################
    
sub get_text_locator {

    my ($data) = @_;
    my %locator;
    my $span;

    ($span) = extract_sgml_span ("charseq", $data);

    ($locator{start}) = extract_sgml_span ("start", $span);
    die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	"No 'start' tag found in data '$span'\n\n" unless defined $locator{start};
    ($locator{end}) = extract_sgml_span ("end", $span);
    die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	"No 'end' tag found in data '$span'\n\n" unless defined $locator{end};
    return {%locator};
}

#################################
    
sub get_audio_locator {

    my ($data) = @_;
    my %locator;
    my ($span, $tstart, $tdur);

    ($span) = extract_sgml_span ("times", $data);

    ($tstart) = extract_sgml_span ("tstart", $span);
    die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	"No 'tstart' tag found in data '$span'\n\n" unless defined $tstart;
    $locator{tstart} = $tstart;
    ($tdur) = extract_sgml_span ("tdur", $span);
    die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	"No 'tdur' tag found in data '$span'\n\n" unless defined $tdur;
    $locator{tdur} = $tdur;
    return {%locator};
}

#################################
    
sub get_image_locator {

    my ($locator_data) = @_;
    my (%box, @boxlist);
    my ($span, $tag, $data, $start, $end, $text);
    
    ($span) = extract_sgml_span ("bbxlist", $locator_data);

    my $nboxes = 0;
    while ($span and (($tag, $data, $span) = 
	   extract_sgml_tag_and_span ("bbx", $span)))
    {
	$box{page} = extract_sgml_tag_attribute ("Signal", $tag);
	($box{x_start}) = extract_sgml_span ("x1", $data);
	($box{y_start}) = extract_sgml_span ("y1", $data);
	($box{width}) = extract_sgml_span ("x2", $data);
	($box{height}) = extract_sgml_span ("y2", $data);

	die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	    "No 'x1' tag found in data '$data'\n\n" unless defined $box{x_start};
	die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	    "No 'y1' tag found in data '$data'\n\n" unless defined $box{y_start};
	die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	    "No 'x2' tag found in data '$data'\n\n" unless defined $box{width};
	die "\n\nFATAL INPUT ERROR for entity '$input_entity' in file '$input_file':\n".
	    "No 'y2' tag found in data '$data'\n\n" unless defined $box{height};

	push @boxlist, {%box};
	$nboxes++;
    }
    die "\n\nFATAL INPUT ERROR for entity '$input_entity' in document '$input_doc'".
	"in file '$input_file':\n  image mention contains no boxes: '$locator_data'\n\n"
	    if ($nboxes eq 0);
    return @boxlist;
}

#################################

sub extract_sgml_span {

    my ($tag_name, $input) = @_;
    my $span;

    unless ($input =~ m|<$tag_name\s*.*?>\s*|s) {return;}
    $span = $';
    unless ($span =~ m|\s*</$tag_name>\s*|s) {return;}
    $span = $`;
    return ($span, $');
}

#################################

sub extract_sgml_tag {

    my ($tag_name, $input) = @_;
    my $tag;

    unless ($input =~ m|<$tag_name\s*|s) {return;}
    $tag = $';
    unless ($tag =~ m|\s*>\s*|) {return;}
    $tag = $`;
}

#################################

sub extract_sgml_tag_and_span {
    
    my ($tag_name, $input) = @_;
    my ($tag, $span);
    
    unless ($input =~ m|<$tag_name\s*|s) {return;}
    $tag = $';
    unless ($tag =~ m|\s*>|) {return;}
    $tag = $`;
    $span = $';
    if ($span =~ m|\s*</$tag_name>\s*|s) {
	$span = $`;
	($tag, $span, $');
    }
    else {
	($tag);
    }
}

#################################

sub extract_sgml_tag_attribute {

    my ($attribute, $tag_name) = @_;
    my $value;

    unless ($tag_name =~ m|$attribute[\s=]+\"|) {return;}
    $value = $';
    unless ($value =~ m|\"|) {return;}
    $value = $`;
}

#################################

sub date_time_stamp {

    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime();
    my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
    my ($date, $time);

    $time = sprintf "%2.2d:%2.2d:%2.2d", $hour, $min, $sec;
    $date = sprintf "%4.4s %3.3s %s", 1900+$year, $months[$mon], $mday;
    return ($date, $time);
}

#################################

sub print_EDT_data {

    my ($doc_id, $doc);
    my ($entity, $mention, $name);
    
    while (@_) {
	$doc_id = shift(@_);
	$doc = shift(@_);
	print "doc ID = '$doc_id', ";
	print "TYPE = '$doc->{TYPE}', ";
	print "FILE = '$doc->{FILE}'\n";
	foreach $entity (@{$doc->{entities}}) {
	    print "  entity ID = '$entity->{ID}', ";
	    print "TYPE = '$entity->{TYPE}'\n";
	    foreach $mention (@{$entity->{mentions}}) {
		print "    mention TYPE = '$mention->{TYPE}', ";
		print "head = '$mention->{head}{text}', ";
		print "extent = '$mention->{extent}{text}'\n";
	    }
	    foreach $name (@{$entity->{names}}) {
		print "      name extent = '$name->{text}'\n";
	    }
	}
    }
}